"""
(C) 2014-2019 Roman Sirokov and contributors
Licensed under BSD license

http://github.com/r0x0r/pywebview/
"""

from __future__ import annotations

import inspect
import json
import logging
import os
import re
import sys
import traceback
import urllib.parse
from collections import UserDict
from glob import glob
from http.cookies import SimpleCookie
from platform import architecture
from threading import Thread
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import webview
from webview.dom import _dnd_state
from webview.errors import WebViewException

if TYPE_CHECKING:
    from webview.window import Window

_TOKEN = uuid4().hex

DEFAULT_HTML = """
    <!doctype html>
    <html>
        <head>
            <meta name="viewport" content="width=device-width, initial-scale=1, maximum-scale=1.0, user-scalable=0">
        </head>
        <body></body>
    </html>
"""

logger = logging.getLogger('pywebview')


class ImmutableDict(UserDict):
    """ "
    A dictionary that does not allow adding new keys or deleting existing ones.
    Only existing keys can be modified.
    """

    def __init__(self, initial_data=None, **kwargs):
        self.data = {}
        if initial_data:
            self.data.update(initial_data)
        if kwargs:
            self.data.update(kwargs)

    def __setitem__(self, key, value):
        if key not in self.data:
            raise KeyError(f"Cannot add new key '{key}'. Only existing keys can be modified.")
        super().__setitem__(key, value)

    def __delitem__(self, key):
        raise KeyError('Deleting keys is not allowed.')


def is_app(url: str | callable | None) -> bool:
    """Returns true if 'url' is a WSGI or ASGI app."""
    return callable(url)


def is_local_url(url: str | callable | None) -> bool:
    return not (
        (is_app(url))
        or (
            (not url)
            or (url.startswith('http://'))
            or (url.startswith('https://'))
            or url.startswith('file://')
        )
    )


def needs_server(urls: list[str]) -> bool:
    return bool([url for url in urls if (is_app(url) or is_local_url(url))])


def get_app_root() -> str:
    """
    Gets the file root of the application.
    """

    if hasattr(sys, '_MEIPASS'):  # Pyinstaller
        return sys._MEIPASS

    if os.getenv('RESOURCEPATH'):  # py2app
        return os.getenv('RESOURCEPATH')

    if getattr(sys, 'frozen', False):  # cx_freeze
        return os.path.dirname(sys.executable)

    if 'pytest' in sys.modules and os.getenv('PYWEBVIEW_TEST'):
        return os.path.join(os.path.dirname(__file__), '..', 'tests')

    if hasattr(sys, 'getandroidapilevel'):
        return os.getenv('ANDROID_APP_PATH')

    return os.path.dirname(os.path.realpath(sys.argv[0]))


def abspath(path: str) -> str:
    """
    Make path absolute, using the application root
    """
    path = os.fspath(path)
    if not os.path.isabs(path):
        path = os.path.join(get_app_root(), path)
    return os.path.normpath(path)


def base_uri(relative_path: str = '') -> str:
    """Get absolute path to resource, works for dev and for PyInstaller"""
    base_path = get_app_root()
    if not os.path.exists(base_path):
        raise ValueError(f'Path {base_path} does not exist')

    return f'file://{os.path.join(base_path, relative_path)}'


def create_cookie(input_: dict[Any, Any] | str) -> SimpleCookie[str]:
    if isinstance(input_, dict):
        cookie = SimpleCookie()
        name = input_['name']
        cookie[name] = input_['value']
        cookie[name]['path'] = input_['path']
        cookie[name]['domain'] = input_['domain']
        cookie[name]['expires'] = input_['expires']
        cookie[name]['secure'] = input_['secure']
        cookie[name]['httponly'] = input_['httponly']

        if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
            cookie[name]['samesite'] = input_.get('samesite')

        return cookie

    if isinstance(input_, str):
        return SimpleCookie(input_)

    raise WebViewException('Unknown input to create_cookie')


def parse_file_type(file_type: str) -> tuple[str, str]:
    """
    :param file_type: file type string 'description (*.file_extension1;*.file_extension2)' as required by file filter in create_file_dialog
    :return: (description, file extensions) tuple
    """
    valid_file_filter = r'^([\w ]+)\((\*(?:\.(?:\w+|\*))*(?:;\*(?:\.(?:\w+|\*))*)*)\)$'
    match = re.search(valid_file_filter, file_type)

    if match:
        return match.group(1).rstrip(), match.group(2)
    raise ValueError(f'{file_type} is not a valid file filter')


def inject_pywebview(platform: str, window: Window) -> str:
    """ "
    Generates and injects a global window.pywebview object. The object contains exposed API functions
    as well as utility functions required by pywebview. The function fires before_load event before
    injecting the object and loaded event after the object is injected.
    """
    exposed_objects = []

    def get_args(func: object):
        params = list(inspect.getfullargspec(func).args)
        return params

    def get_functions(obj: object, base_name: str = '', functions: dict[str, object] = None):
        obj_id = id(obj)
        if obj_id in exposed_objects:
            return functions
        else:
            exposed_objects.append(obj_id)

        if functions is None:
            functions = {}

        for name in dir(obj):
            try:
                full_name = f'{base_name}.{name}' if base_name else name
                if name.startswith('_'):
                    continue

                attr = getattr(obj, name)
                if not getattr(attr, '_serializable', True):
                    continue

                if inspect.ismethod(attr) or inspect.isfunction(attr):
                    functions[full_name] = get_args(attr)[1:]
                # If the attribute is a class or a non-callable object, make a recursive call
                elif inspect.isclass(attr) or (
                    isinstance(attr, object) and not callable(attr) and hasattr(attr, '__module__')
                ):
                    get_functions(attr, full_name, functions)
            except Exception as e:
                logger.error(f'Error while processing {full_name}: {e}')
                continue

        return functions

    def generate_func():
        functions = get_functions(window._js_api)

        if len(window._functions) > 0:
            expose_functions = {name: get_args(f) for name, f in window._functions.items()}
        else:
            expose_functions = {}

        functions.update(expose_functions)

        return [{'func': name, 'params': params} for name, params in functions.items()]

    def generate_js_object():
        window.run_js(js_code)
        logger.debug('_pywebviewready event fired')

        try:
            with window._expose_lock:
                func_list = generate_func()
                window.run_js(finish_script % {'functions': json.dumps(func_list)})
                window.events._pywebviewready.set()
                window.events.loaded.set()
                logger.debug('loaded event fired')
        except Exception as e:
            logger.exception(e)
            window.events.loaded.set()

    window.events.before_load.set()
    logger.debug('before_load event fired. injecting pywebview object')
    js_code, finish_script = load_js_files(window, platform)
    thread = Thread(target=generate_js_object)
    thread.start()


def js_bridge_call(window: Window, func_name: str, param: Any, value_id: str) -> None:
    """
    Calls a function from the JS API and executes it in Python. The function is executed in a separate
    thread to prevent blocking the UI thread. The result is then passed back to the JS API.
    """

    def _call():
        try:
            result = func(*func_params)
            result = json.dumps(result).replace('\\', '\\\\').replace("'", "\\'")
            retval = f"{{value: '{result}'}}"
        except Exception as e:
            logger.error(traceback.format_exc())
            error = {'message': str(e), 'name': type(e).__name__, 'stack': traceback.format_exc()}
            result = json.dumps(error).replace('\\', '\\\\').replace("'", "\\'")
            retval = f"{{isError: true, value: '{result}'}}"

        window.evaluate_js(
            f'window.pywebview._returnValuesCallbacks["{func_name}"]["{value_id}"]({retval})'
        )

    def get_nested_attribute(obj: object, attr_str: str):
        attributes = attr_str.split('.')
        for attr in attributes:
            obj = getattr(obj, attr, None)
            if obj is None:
                return None
        return obj

    if func_name == 'pywebviewMoveWindow':
        window.move(*param)
        return

    if func_name == 'pywebviewEventHandler':
        event = param['event']
        node_id = param['nodeId']
        element = window.dom._elements.get(node_id)

        if not element:
            return

        if event['type'] == 'drop':
            files = event['dataTransfer'].get('files', [])
            for file in files:
                path = [
                    item
                    for item in _dnd_state['paths']
                    if urllib.parse.unquote(item[0]) == file['name']
                ]
                if len(path) == 0:
                    continue

                file['pywebviewFullPath'] = urllib.parse.unquote(path[0][1])
                _dnd_state['paths'].remove(path[0])

        for handler in element._event_handlers.get(event['type'], []):
            thread = Thread(target=handler, args=(event,))
            thread.start()

        return

    if func_name == 'pywebviewAsyncCallback':
        value = json.loads(param) if param is not None else None

        if callable(window._callbacks[value_id]):
            window._callbacks[value_id](value)
        else:
            logger.error(
                f'Async function executed and callback is not callable. Returned value {value}'
            )

        del window._callbacks[value_id]
        return

    if func_name == 'pywebviewStateUpdate':
        window.state.__setattr__(param['key'], param['value'], False)
        return

    if func_name == 'pywebviewStateDelete':
        special_key = '__pywebviewHaltUpdate__' + param
        delattr(window.state, special_key)
        return

    func = window._functions.get(func_name) or get_nested_attribute(window._js_api, func_name)

    if func is not None:
        try:
            func_params = param
            thread = Thread(target=_call)
            thread.start()
        except Exception:
            logger.exception('Error occurred while evaluating function %s', func_name)
    else:
        logger.error('Function %s() does not exist', func_name)


def load_js_files(window: Window, platform: str) -> str:
    """
    Load JS files in the order they should be loaded.
    The order is polyfill, api, the rest and finish.js.
    Return the concatenated JS code and the finish script, which must be loaded last and
    separately in order to
    """
    js_dir = get_js_dir()
    logger.debug('Loading JS files from %s', js_dir)
    js_files = glob(os.path.join(js_dir, '**', '*.js'), recursive=True)
    ordered_js_files = sort_js_files(js_files)
    js_code = ''
    finish_script = ''

    for file in ordered_js_files:
        with open(file) as f:
            name = os.path.splitext(os.path.basename(file))[0]
            content = f.read()
            params = {}

            if name == 'api':
                params = {
                    'token': _TOKEN,
                    'platform': platform,
                    'uid': window.uid,
                    'js_api_endpoint': window.js_api_endpoint,
                }
            elif name == 'customize':
                params = {
                    'text_select': str(window.text_select),
                    'drag_selector': webview.settings['DRAG_REGION_SELECTOR'],
                    'drag_region_direct_target_only': str(
                        webview.settings['DRAG_REGION_DIRECT_TARGET_ONLY']
                    ),
                    'zoomable': str(window.zoomable),
                    'draggable': str(window.draggable),
                    'easy_drag': str(
                        platform == 'edgechromium' and window.easy_drag and window.frameless
                    ),
                }
            elif name == 'state':
                params = {'state': json.dumps(window.state)}
            elif name == 'finish':
                finish_script = content
                continue
            elif name == 'polyfill' and platform != 'mshtml':
                continue

            js_code += content % params

    return js_code, finish_script


def get_js_dir() -> str:
    """
    Get the path to the directory with Javascript files.
    """
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'js')

    if os.path.exists(path):
        return path

    # try py2app frozen path. This is hacky, but it works.
    # See https://github.com/r0x0r/pywebview/issues/1565
    if '.zip' in path:
        base_path = path.split('.zip')[0]
        dir = os.path.dirname(base_path)

        for file in os.listdir(dir):
            if file.startswith('python') and os.path.isdir(os.path.join(dir, file)):
                js_path = os.path.join(dir, file, 'webview', 'js')
                if os.path.exists(js_path):
                    return js_path

    raise FileNotFoundError(f'Cannot find JS directory in {path}')


def sort_js_files(js_files: list[str]) -> list[str]:
    """
    Sorts JS files in the order they should be loaded. Polyfill first, then API, then the rest and
    finally finish.js that fires a pywebviewready event.
    """
    LOAD_ORDER = {'polyfill': 0, 'api': 1, 'state': 2}

    ordered_js_files = []
    remaining_js_files = []

    for file in js_files:
        basename = os.path.splitext(os.path.basename(file))[0]
        if basename not in LOAD_ORDER:
            ordered_js_files.append(file)
        else:
            remaining_js_files.append((basename, file))

    for basename, file in sorted(remaining_js_files, key=lambda x: LOAD_ORDER[x[0]]):
        ordered_js_files.insert(LOAD_ORDER[basename], file)

    return ordered_js_files


def escape_string(string: str) -> str:
    return (
        string.replace('\\', '\\\\')
        .replace('"', r'\"')
        .replace('\n', r'\n')
        .replace('\r', r'\r')
        .replace("'", r'\'')
    )


def inject_base_uri(content: str, base_uri: str) -> str:
    pattern = r'<%s(?:[\s]+[^>]*|)>'
    base_tag = f'<base href="{base_uri}">'

    match = re.search(pattern % 'base', content)

    if match:
        return content

    match = re.search(pattern % 'head', content)
    if match:
        tag = match.group()
        return content.replace(tag, tag + base_tag)

    match = re.search(pattern % 'html', content)
    if match:
        tag = match.group()
        return content.replace(tag, tag + base_tag)

    match = re.search(pattern % 'body', content)
    if match:
        tag = match.group()
        return content.replace(tag, base_tag + tag)

    return base_tag + content


def interop_dll_path(dll_name: str) -> str:
    if dll_name == 'WebBrowserInterop.dll':
        dll_name = (
            'WebBrowserInterop.x64.dll'
            if architecture()[0] == '64bit'
            else 'WebBrowserInterop.x86.dll'
        )

    # Unfrozen path
    dll_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'lib', dll_name)
    if os.path.exists(dll_path):
        return dll_path

    dll_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), 'lib', 'runtimes', dll_name, 'native'
    )
    if os.path.exists(dll_path):
        return dll_path

    # Frozen path, dll in the same dir as the executable
    dll_path = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), dll_name)
    if os.path.exists(dll_path):
        return dll_path

    try:
        # Frozen path packed as onefile
        if hasattr(sys, '_MEIPASS'):  # Pyinstaller
            dll_path = os.path.join(sys._MEIPASS, dll_name)

        elif getattr(sys, 'frozen', False):  # cx_freeze
            dll_path = os.path.join(sys.executable, dll_name)

        if os.path.exists(dll_path):
            return dll_path
    except Exception:
        pass

    raise FileNotFoundError(f'Cannot find {dll_name}')


def environ_append(key: str, *values: str, sep=' ') -> None:
    """Append values to an environment variable, separated by sep"""
    values = list(values)

    existing = os.environ.get(key, '')
    if existing:
        values = [existing] + values

    os.environ[key] = sep.join(values)


def css_to_camel(css_case_string: str) -> str:
    words = css_case_string.split('-')
    camel_case_string = words[0] + ''.join(word.capitalize() for word in words[1:])
    return camel_case_string


def android_jar_path() -> str:
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'lib', 'pywebview-android.jar')


def stringify_headers(headers: dict[str, Any]) -> dict[str, str]:
    return {k: str(v) for k, v in headers.items()}



from android.runnable import run_on_ui_thread  # noqa
from android.activity import _activity as act  # noqa
from webview.platforms.android.base import EventLoop
from webview.platforms.android.event import EventDispatcher


class App(EventDispatcher):
    # Return the current running App instance
    _running_app = None

    def __init__(self, **kwargs):
        App._running_app = self
        super().__init__(**kwargs)
        self._eventloop = EventLoop()

        #: The *root* widget returned by the :meth:`build_view`
        self.root = None

    def build_view(self):
        """Initializes the application; it will be called only once.
        If this method returns a widget (tree), it will be used as the root
        widget and added to the window.

        :return:
            None or a root :class:`~android.widget.RelativeLayout` instance
            if no self.root exists."""
        return self.root

    @run_on_ui_thread
    def set_content_view(self, root):
        act.setContentView(root)

    def on_create(self, activity, saved_instance_state):
        """Event handler for the `on_create` event which is fired after
        initialization (after build() has been called) but before the
        application has started running.
        """

    def on_pause(self, activity):
        """Event handler called when Pause mode is requested"""

    def on_destroy(self, activity):
        """Event handler for the `on_destroy` event which is fired when the
        application has finished running (i.e. the window is about to be
        closed).
        """

    def on_resume(self, activity):
        pass

    def on_start(self, activity):
        pass

    def on_stop(self, activity):
        pass

    def run(self):
        self.build_view()
        self._eventloop.status = 'created'
        self._eventloop.mainloop()

    @staticmethod
    def get_running_app():
        """Return the currently running application instance."""
        return App._running_app

    def pause(self):
        """Pause the application.

        On Android set OS state to pause, Kivy app state follows.
        No functionality on other OS.
        .. versionadded:: 2.2.0
        """
        act.moveTaskToBack(True)

    def stop(self):
        self.dispatch('on_destroy')
        self._eventloop.close()
        App._running_app = None



class ClassList:
    def __init__(self, element, classes=None):
        self.__element = element

        if classes:
            classes = ' '.join(classes)
            self.__element._window.evaluate_js(
                f"{self.__element._query_command}; element.className = '{classes}'"
            )

    def append(self, cls):
        self.__element._window.run_js(
            f"{self.__element._query_command}; element.classList.add('{cls}')"
        )

    def remove(self, cls):
        self.__element._window.run_js(
            f"{self.__element._query_command}; element.classList.remove('{cls}')"
        )

    def toggle(self, cls):
        self.__element._window.run_js(
            f"{self.__element._query_command}; element.classList.toggle('{cls}')"
        )

    def __get_classes(self):
        classes = self.__element._window.evaluate_js(
            f'{self.__element._query_command}; element.className'
        ).split(' ')
        return [c for c in classes if c != '']

    def __getitem__(self, index):
        classes = self.__get_classes()
        return classes[index]

    def __len__(self):
        classes = self.__get_classes()
        return len(classes)

    def __str__(self):
        classes = self.__get_classes()
        return str(classes)

    def __repr__(self):
        classes = self.__get_classes()
        return repr(classes)

    def clear(self):
        self.__element._window.run_js(f"{self.__element._query_command}; element.className = ''")



class Screen:
    def __init__(
        self, x: int, y: int, width: int, height: int, frame: object = None, scale: float = 1.0
    ) -> None:
        self.x = int(x)
        self.y = int(y)
        self.width = int(width)
        self.height = int(height)
        self.frame = frame
        self.scale = float(scale)

    @property
    def physical_x(self) -> int:
        """X coordinate in physical pixels."""
        return int(self.x * self.scale)

    @property
    def physical_y(self) -> int:
        """Y coordinate in physical pixels."""
        return int(self.y * self.scale)

    @property
    def physical_width(self) -> int:
        """Width in physical pixels."""
        return int(self.width * self.scale)

    @property
    def physical_height(self) -> int:
        """Height in physical pixels."""
        return int(self.height * self.scale)

    @property
    def dpi(self) -> int:
        """DPI (dots per inch) for this screen."""
        return int(self.scale * 96)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        scale_str = f' {self.scale:.2f}x' if self.scale != 1.0 else ''
        return f'{self.width}x{self.height} at {self.x},{self.y}{scale_str}'



from enum import Enum
from typing import Callable


class ManipulationMode(Enum):
    LastChild = 'LAST_CHILD'
    FirstChild = 'FIRST_CHILD'
    Before = 'BEFORE'
    After = 'AFTER'
    Replace = 'REPLACE'


class DOMEventHandler:
    def __init__(
        self,
        callback: Callable,
        prevent_default: bool = False,
        stop_propagation: bool = False,
        stop_immediate_propagation: bool = False,
        debounce: int = 0,
    ):
        self.__callback = callback
        self.__prevent_default = prevent_default
        self.__stop_propagation = stop_propagation
        self.__stop_immediate_propagation = stop_immediate_propagation
        self.__debounce = debounce

    @property
    def callback(self):
        return self.__callback

    @property
    def prevent_default(self):
        return self.__prevent_default

    @property
    def stop_propagation(self):
        return self.__stop_propagation

    @property
    def stop_immediate_propagation(self):
        return self.__stop_immediate_propagation

    @property
    def debounce(self):
        return self.__debounce


_dnd_state = {'num_listeners': 0, 'paths': []}



import json
from threading import Thread
from typing import Any, Callable

from typing_extensions import Self

from webview.util import escape_string

try:
    from enum import StrEnum  # Python 3.11 and above
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


class StateEventType(StrEnum):
    CHANGE = 'change'
    DELETE = 'delete'


class State(dict):
    _serializable = False

    def __init__(self, window):
        self.__event_handlers = []
        self.__window = window

    def __update_js(self, key, value):
        special_key = '__pywebviewHaltUpdate__' + key
        script = f"window.pywebview.state.{special_key} = JSON.parse('{escape_string(json.dumps(value))}')"
        self.__window.run_js(script)

    def __notify_handlers(self, type, key, value=None):
        def notify_handlers():
            for handler in self.__event_handlers:
                if callable(handler):
                    if type == StateEventType.CHANGE:
                        handler(type, key, value)
                    elif type == StateEventType.DELETE:
                        handler(type, key, value)

        t = Thread(target=notify_handlers)
        t.start()

    def _set_state_value(self, key, value, should_update_js=True):
        """Internal method to set state values, used by both __setattr__ and __setitem__"""

        def update_and_notify():
            self.__update_js(key, value)
            self.__notify_handlers(StateEventType.CHANGE, key, value)

        if key not in self or super().get(key) != value:
            super().__setitem__(key, value)

            if not should_update_js:
                self.__notify_handlers(StateEventType.CHANGE, key, value)
                return

            if self.__window.events.loaded.is_set():
                update_and_notify()
            else:
                self.__window.events.loaded += update_and_notify

    def __setattr__(self, key, value, should_update_js=True):
        if key.startswith('_State__'):
            super().__setattr__(key, value)
            return

        self._set_state_value(key, value, should_update_js)

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError(f"'{type(self).__key__}' object has no attribute '{key}'")

    def _delete_state_value(self, key, should_update_js=True):
        """Internal method to delete state values, used by both __delattr__ and __delitem__"""
        if key in self:
            old_value = self[key]
            super().__delitem__(key)

            if should_update_js:
                special_key = '__pywebviewHaltUpdate__' + key
                self.__window.run_js(f'delete window.pywebview.state.{special_key}')

            self.__notify_handlers(StateEventType.DELETE, key, old_value)
            return old_value
        return None

    def __delattr__(self, key):
        if key.startswith('__pywebviewHaltUpdate__'):
            key = key.replace('__pywebviewHaltUpdate__', '')
            halt_update = True
        else:
            halt_update = False

        old_value = self._delete_state_value(key, should_update_js=not halt_update)
        if old_value is None:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __getitem__(self, key):
        """Support dictionary-style access: state['key']"""
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        """Support dictionary-style assignment: state['key'] = value"""
        self._set_state_value(key, value, should_update_js=True)

    def __delitem__(self, key):
        """Support dictionary-style deletion: del state['key']"""
        old_value = self._delete_state_value(key, should_update_js=True)
        if old_value is None:
            raise KeyError(key)

    def __add__(self, item: Callable[..., Any]) -> Self:
        self.__event_handlers.append(item)
        return self

    def __sub__(self, item: Callable[..., Any]) -> Self:
        self.__event_handlers.remove(item)
        return self

    def __iadd__(self, item: Callable[..., Any]) -> Self:
        self.__event_handlers.append(item)
        return self

    def __isub__(self, item: Callable[..., Any]) -> Self:
        self.__event_handlers.remove(item)
        return self



import json
from typing import List, Optional, Union

from webview.dom import ManipulationMode
from webview.dom.element import Element


class DOM:
    _serializable = False

    def __init__(self, window):
        self.__window = window
        window.events.loaded += self.__on_loaded
        self._elements = {}

    def __on_loaded(self):
        self._elements = {}

    @property
    def body(self) -> Element:
        return self._elements.get('body', Element(self.__window, 'body'))

    @property
    def document(self) -> Element:
        return self._elements.get('document', Element(self.__window, 'document'))

    @property
    def window(self) -> Element:
        return self._elements.get('window', Element(self.__window, 'window'))

    def create_element(
        self,
        html: str,
        parent: Optional[Union[Element, str]] = None,
        mode: ManipulationMode = ManipulationMode.LastChild,
    ) -> Element:
        self.__window.events.loaded.wait()

        if isinstance(parent, Element):
            parent_command = parent._query_command
        elif isinstance(parent, str):
            parent_command = f'var element = document.querySelector("{parent}");'
        else:
            parent_command = 'var element = document.body;'
        if not isinstance(html, str):
            html = str(html)

        node_id = self.__window.evaluate_js(
            f"""
            {parent_command};
            var template = document.createElement('template');
            template.innerHTML = {json.dumps(html)}.trim();
            var newElement = template.content.firstChild;
            pywebview._insertNode(newElement, element, '{mode.value}')
            pywebview._getNodeId(newElement);
        """
        )

        return Element(self.__window, node_id)

    def get_element(self, selector: str) -> Optional[Element]:
        node_id = self.__window.evaluate_js(
            f"""
            var element = document.querySelector('{selector}');
            pywebview._getNodeId(element);
        """
        )

        return Element(self.__window, node_id) if node_id else None

    def get_elements(self, selector: str) -> List[Element]:
        code = f"""
            var elements = document.querySelectorAll('{selector}');
            nodeIds = [];
            for (var i = 0; i < elements.length; i++) {{
                var nodeId = pywebview._getNodeId(elements[i]);
                nodeIds.push(nodeId);
            }}

            nodeIds
        """

        node_ids = self.__window.evaluate_js(code)
        return [Element(self.__window, node_id) for node_id in node_ids]



import json
import logging
import os
import platform
import signal
import socket
import sys
import webbrowser
from copy import copy, deepcopy
from functools import partial
from threading import Event, Semaphore, Thread
from uuid import uuid1

from qtpy import PYQT6, PYSIDE6, QtCore
from qtpy.QtCore import QByteArray, QJsonValue
from qtpy.QtGui import QColor, QIcon, QScreen
from qtpy.QtWidgets import QAction, QApplication, QFileDialog, QMainWindow, QMenuBar, QMessageBox

from webview import FileDialog, _state, settings, windows
from webview.dom import _dnd_state
from webview.menu import Menu, MenuAction, MenuSeparator
from webview.models import Request
from webview.screen import Screen
from webview.util import (
    DEFAULT_HTML,
    create_cookie,
    environ_append,
    inject_pywebview,
    js_bridge_call,
)
from webview.window import FixPoint, Window

try:
    from qtpy.QtNetwork import QSslCertificate, QSslConfiguration
    from qtpy.QtWebChannel import QWebChannel
    from qtpy.QtWebEngineCore import QWebEngineUrlRequestInterceptor
    from qtpy.QtWebEngineWidgets import QWebEnginePage as QWebPage
    from qtpy.QtWebEngineWidgets import QWebEngineProfile, QWebEngineSettings
    from qtpy.QtWebEngineWidgets import QWebEngineView as QWebView

    renderer = 'qtwebengine'
    is_webengine = True
except ImportError:
    from PyQt5.QtNetwork import QSslCertificate, QSslConfiguration
    from PyQt5.QtWebKitWidgets import QWebPage, QWebView

    is_webengine = False
    renderer = 'qtwebkit'

logger = logging.getLogger('pywebview')
logger.debug(f'Using Qt {QtCore.__version__}')

if is_webengine and QtCore.QSysInfo.productType() in ['arch', 'manjaro', 'nixos', 'rhel', 'pop']:
    # I don't know why, but it's a common solution for #890 (White screen displayed)
    # such as:
    # - https://github.com/LCA-ActivityBrowser/activity-browser/pull/954/files
    # - https://bugs.archlinux.org/task/73957
    # - https://www.google.com/search?q=arch+rstudio+no+sandbox
    # And sometimes it needs two "--no-sandbox" flags

    environ_append('QTWEBENGINE_CHROMIUM_FLAGS', '--no-sandbox', '--no-sandbox')
    logger.debug('Enable --no-sandbox flag for arch/manjaro/nixos')

_app_menu = None
_main_window_created = Event()
_main_window_created.clear()

# suppress invalid style override error message on some Linux distros
os.environ['QT_STYLE_OVERRIDE'] = ''
_qt6 = True if PYQT6 or PYSIDE6 else False
_profile_storage_path = _state['storage_path'] or os.path.join(
    os.path.expanduser('~'), '.pywebview'
)


def _sigint_handler(signum, frame):
    """
    Handler for SIGINT signal (Ctrl+C).

    Qt applications don't handle SIGINT by default because Python signal handlers
    can only be called when the Python interpreter is running. Since the Qt event
    loop runs in C++, signals are noted but handlers aren't called until Python
    code executes. The timer in create_window() ensures the Python interpreter
    runs periodically to process pending signals.

    See: https://stackoverflow.com/a/4939113
    """
    global _app
    if _app:
        _app.quit()


class BrowserView(QMainWindow):
    instances = {}
    inspector_port = None  # The localhost port at which the Remote debugger listens

    global_menubar_other_objects = []
    global_menubar_top_menus = []

    create_window_trigger = QtCore.Signal(object)
    set_title_trigger = QtCore.Signal(str)
    load_url_trigger = QtCore.Signal(str)
    html_trigger = QtCore.Signal(str, str)
    confirmation_dialog_trigger = QtCore.Signal(str, str, str)
    file_dialog_trigger = QtCore.Signal(int, str, bool, str, str)
    destroy_trigger = QtCore.Signal()
    hide_trigger = QtCore.Signal()
    show_trigger = QtCore.Signal()
    fullscreen_trigger = QtCore.Signal()
    window_size_trigger = QtCore.Signal(int, int, FixPoint)
    window_move_trigger = QtCore.Signal(int, int)
    window_maximize_trigger = QtCore.Signal()
    window_minimize_trigger = QtCore.Signal()
    window_restore_trigger = QtCore.Signal()
    current_url_trigger = QtCore.Signal()
    evaluate_js_trigger = QtCore.Signal(str, str)
    on_top_trigger = QtCore.Signal(bool)

    class JSBridge(QtCore.QObject):
        qtype = QtCore.QJsonValue if is_webengine else str

        def __init__(self, parent):
            super(BrowserView.JSBridge, self).__init__()
            self._parent = parent
            self.window = parent.pywebview_window

        @QtCore.Slot(str, qtype, str, result=str)
        def call(self, func_name, param, value_id):
            func_name = BrowserView._convert_string(func_name)
            param = BrowserView._convert_string(param)

            if func_name == '_pywebviewAlert':
                QMessageBox.information(self._parent, 'Message', str(param))
            else:
                return js_bridge_call(self.window, func_name, json.loads(param), value_id)

    class WebView(QWebView):
        def __init__(self, parent=None):
            super(BrowserView.WebView, self).__init__(parent)

            if parent.frameless and parent.easy_drag:
                QApplication.instance().installEventFilter(self)
                self.setMouseTracking(True)

            self.transparent = parent.transparent
            if parent.transparent:
                self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
                self.setAttribute(QtCore.Qt.WA_OpaquePaintEvent, False)
                self.setStyleSheet('background: transparent;')

        def contextMenuEvent(self, event):
            if _qt6:
                menu = self.createStandardContextMenu()
            else:
                menu = self.page().createStandardContextMenu()

            # If 'Inspect Element' is present in the default context menu, it
            # means the inspector is already up and running.
            for i in menu.actions():
                if i.text() == 'Inspect Element':
                    break
            else:
                # Inspector is not up yet, so create a pseudo 'Inspect Element'
                # menu that will fire it up.
                inspect_element = QAction('Inspect Element', menu)
                inspect_element.triggered.connect(self.show_inspector)
                menu.addAction(inspect_element)

            menu.exec_(event.globalPos())

        def dragEnterEvent(self, e):
            if e.mimeData().hasUrls and _dnd_state['num_listeners'] > 0:
                e.acceptProposedAction()

            return super().dragEnterEvent(e)

        def dragMoveEvent(self, e):
            if e.mimeData().hasUrls and _dnd_state['num_listeners'] > 0:
                e.acceptProposedAction()

            return super().dragMoveEvent(e)

        def dropEvent(self, e):
            if e.mimeData().hasUrls and _dnd_state['num_listeners'] > 0:
                files = [
                    (os.path.basename(value.toString()), value.toString().replace('file://', ''))
                    for value in e.mimeData().urls()
                    if value.toString().startswith('file://')
                ]
                _dnd_state['paths'] += files

            return super().dropEvent(e)

        # Create a new webview window pointing at the Remote debugger server
        def show_inspector(self):
            uid = self.parent().uid + '-inspector'
            try:
                # If inspector already exists, bring it to the front
                BrowserView.instances[uid].raise_()
                BrowserView.instances[uid].activateWindow()
            except KeyError:
                title = f'Web Inspector - {self.parent().title}'
                url = f'http://localhost:{BrowserView.inspector_port}'
                window = Window('web_inspector', title, url, '', 700, 500)
                window.localization = self.parent().localization

                inspector = BrowserView(window)
                inspector.show()

        def mousePressEvent(self, event):
            if event.button() == QtCore.Qt.LeftButton:
                self.drag_pos = event.globalPos() - self.parent().frameGeometry().topLeft()

            event.accept()

        def mouseMoveEvent(self, event):
            parent = self.parent()
            if (
                parent.frameless and parent.easy_drag and event.buttons() & QtCore.Qt.LeftButton
            ):  # left button is pressed
                parent.move(event.globalPos() - self.drag_pos)

        def eventFilter(self, object, event):
            if object.parent() == self:
                if event.type() == QtCore.QEvent.MouseMove:
                    self.mouseMoveEvent(event)
                elif event.type() == QtCore.QEvent.MouseButtonPress:
                    self.mousePressEvent(event)

            return False

    class RequestInterceptor(QWebEngineUrlRequestInterceptor):
        def __init__(self, window):
            super().__init__()
            self.window = window

        def interceptRequest(self, info):
            if len(self.window.events.request_sent) == 0:
                return

            url = info.requestUrl().toString()
            method = info.requestMethod()

            if 'httpHeaders' in dir(info):
                headers = {
                    k.data().decode('utf-8'): k.data().decode('utf-8')
                    for k, v in info.httpHeaders().items()
                }
            else:
                headers = {}

            request = Request(url, method, headers)
            self.window.events.request_sent.set(request)

            if request.headers != headers:
                for key, value in request.headers.items():
                    info.setHttpHeader(
                        QByteArray(key.encode('utf-8')), QByteArray(value.encode('utf-8'))
                    )

    # New-window-requests handler for Qt 5.5+ only
    class NavigationHandler(QWebPage):
        def __init__(self, page):
            super().__init__(page.profile())
            self._parent = page.parent

        def acceptNavigationRequest(self, url, type, is_main_frame):
            if settings['OPEN_EXTERNAL_LINKS_IN_BROWSER']:
                webbrowser.open(url.toString(), 2, True)
                return False
            else:
                self._parent.load_url(url.toString())
                return True

    class WebPage(QWebPage):
        def __init__(self, parent=None, profile=None):
            if is_webengine and profile:
                super(BrowserView.WebPage, self).__init__(profile, parent.webview)
            else:
                super(BrowserView.WebPage, self).__init__(parent.webview)

            if is_webengine:
                self.featurePermissionRequested.connect(self.onFeaturePermissionRequested)
                self.nav_handler = BrowserView.NavigationHandler(self)
            else:
                self.nav_handler = None

            if parent.transparent:
                self.setBackgroundColor(QtCore.Qt.transparent)

        if is_webengine:

            def onFeaturePermissionRequested(self, url, feature):
                if feature in (
                    QWebPage.Feature.MediaAudioCapture,
                    QWebPage.Feature.MediaVideoCapture,
                    QWebPage.Feature.MediaAudioVideoCapture,
                ):
                    self.setFeaturePermission(url, feature, 1)  # QWebPage.PermissionGrantedByUser
                else:
                    self.setFeaturePermission(url, feature, 2)  # QWebPage.PermissionDeniedByUser
        else:

            def acceptNavigationRequest(self, frame, request, type):
                if frame is None:
                    webbrowser.open(request.url().toString(), 2, True)
                    return False
                return True

        def userAgentForUrl(self, url):
            user_agent = _state['user_agent']
            if user_agent:
                return user_agent
            else:
                return super().userAgentForUrl(url)

        def createWindow(self, type):
            return self.nav_handler

    def __init__(self, window):
        super().__init__()
        BrowserView.instances[window.uid] = self
        self.uid = window.uid
        self.pywebview_window = window
        self.pywebview_window.native = self

        self.js_bridge = BrowserView.JSBridge(self)

        self.is_fullscreen = False

        self._file_name_semaphore = Semaphore(0)
        self._current_url_semaphore = Semaphore(0)

        self.localization = window.localization

        if window.screen:
            if _qt6:
                self.screen = QScreen.geometry(window.screen.frame)
            else:
                self.screen = window.screen.frame.geometry()
        else:
            if _qt6:
                self.screen = QScreen.geometry(QApplication.primaryScreen())
            else:
                self.screen = QApplication.primaryScreen().geometry()

        self._js_results = {}
        self._current_url = None
        self._file_name = None
        self._confirmation_dialog_results = {}

        self.resize(window.initial_width, window.initial_height)
        self.title = window.title
        self.setWindowTitle(window.title)

        # Set window background color
        self.background_color = QColor()
        self.background_color.setNamedColor(window.background_color)
        palette = self.palette()
        palette.setColor(self.backgroundRole(), self.background_color)
        self.setPalette(palette)

        if not window.resizable:
            self.setFixedSize(window.initial_width, window.initial_height)

        self.setMinimumSize(window.min_size[0], window.min_size[1])

        self.frameless = window.frameless
        self.easy_drag = window.easy_drag
        flags = self.windowFlags()
        if self.frameless:
            flags = flags | QtCore.Qt.FramelessWindowHint

        if window.on_top:
            flags = flags | QtCore.Qt.WindowStaysOnTopHint

        if not window.focus:
            self.setAttribute(QtCore.Qt.WA_ShowWithoutActivating)
            flags = flags | QtCore.Qt.WindowDoesNotAcceptFocus

        self.setWindowFlags(flags)
        self.setAcceptDrops(True)

        self.transparent = window.transparent
        if self.transparent:
            # Override the background color
            self.background_color = QColor('transparent')
            palette = self.palette()
            palette.setColor(self.backgroundRole(), self.background_color)
            self.setPalette(palette)
            # Enable the transparency hint
            self.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self.webview = BrowserView.WebView(self)

        if is_webengine:
            environ_append(
                'QTWEBENGINE_CHROMIUM_FLAGS',
                '--use-fake-ui-for-media-stream',
                '--enable-features=AutoplayIgnoreWebAudio',
            )

        if _state['debug'] and is_webengine:
            # Initialise Remote debugging (need to be done only once)
            if not BrowserView.inspector_port:
                BrowserView.inspector_port = BrowserView._get_debug_port()
                os.environ['QTWEBENGINE_REMOTE_DEBUGGING'] = BrowserView.inspector_port
        else:
            self.webview.setContextMenuPolicy(
                QtCore.Qt.NoContextMenu
            )  # disable right click context menu

        self.cookies = {}

        if is_webengine:
            self.request_interceptor = BrowserView.RequestInterceptor(self.pywebview_window)

            if _state['private_mode']:
                self.profile = QWebEngineProfile()
            else:
                self.profile = QWebEngineProfile('pywebview')
                self.profile.setPersistentStoragePath(_profile_storage_path)

            user_agent = _state['user_agent']
            if user_agent:
                self.profile.setHttpUserAgent(user_agent)

            cookie_store = self.profile.cookieStore()
            cookie_store.cookieAdded.connect(self.on_cookie_added)
            cookie_store.cookieRemoved.connect(self.on_cookie_removed)

            self.profile.setUrlRequestInterceptor(self.request_interceptor)
            self.webview.setPage(BrowserView.WebPage(self, profile=self.profile))
        elif not is_webengine and not _state['private_mode']:
            logger.warning('qtwebkit does not support private_mode')

        user_agent = _state['user_agent']
        # if user_agent and is_webengine:
        #     self.profile.setHttpUserAgent(user_agent)

        if is_webengine:
            self.profile.settings().setAttribute(
                QWebEngineSettings.WebAttribute.LocalContentCanAccessFileUrls,
                settings['ALLOW_FILE_URLS'],
            )

        self.webview.page().loadFinished.connect(self.on_load_finished)

        if settings['ALLOW_DOWNLOADS']:
            self.webview.page().profile().downloadRequested.connect(self.on_download_requested)

        self.setCentralWidget(self.webview)

        self.create_window_trigger.connect(BrowserView.on_create_window)
        self.load_url_trigger.connect(self.on_load_url)
        self.html_trigger.connect(self.on_load_html)
        self.confirmation_dialog_trigger.connect(self.on_confirmation_dialog)
        self.file_dialog_trigger.connect(self.on_file_dialog)
        self.destroy_trigger.connect(self.on_destroy_window)
        self.show_trigger.connect(self.on_show_window)
        self.hide_trigger.connect(self.on_hide_window)
        self.fullscreen_trigger.connect(self.on_fullscreen)
        self.window_size_trigger.connect(self.on_window_size)
        self.window_move_trigger.connect(self.on_window_move)
        self.window_maximize_trigger.connect(self.on_window_maximize)
        self.window_minimize_trigger.connect(self.on_window_minimize)
        self.window_restore_trigger.connect(self.on_window_restore)
        self.current_url_trigger.connect(self.on_current_url)
        self.evaluate_js_trigger.connect(self.on_evaluate_js)
        self.set_title_trigger.connect(self.on_set_title)
        self.on_top_trigger.connect(self.on_set_on_top)

        if is_webengine and platform.system() != 'OpenBSD':
            self.channel = QWebChannel(self.webview.page())
            self.webview.page().setWebChannel(self.channel)

        if window.fullscreen:
            self.toggle_fullscreen()

        if window.real_url is not None:
            self.webview.setUrl(QtCore.QUrl(window.real_url))
        elif window.uid == 'web_inspector':
            self.webview.setUrl(QtCore.QUrl(window.original_url))
        elif window.html:
            self.webview.setHtml(window.html, QtCore.QUrl(''))
        else:
            self.webview.setHtml(DEFAULT_HTML, QtCore.QUrl(''))

        if window.initial_x is not None and window.initial_y is not None:
            self.move(self.screen.x() + window.initial_x, self.screen.y() + window.initial_y)
        else:
            offset = -16 if _qt6 else 0
            center = self.screen.center() - self.rect().center()
            self.move(center.x(), center.y() + offset)

        if not window.minimized:
            self.activateWindow()
            self.raise_()

        if _state['icon']:
            icon = QIcon(_state['icon'])
            self.setWindowIcon(icon)

        self.pywebview_window.events.before_show.set()

    def on_set_title(self, title):
        self.setWindowTitle(title)

    def on_confirmation_dialog(self, title, message, uuid):
        uuid_ = BrowserView._convert_string(uuid)
        reply = QMessageBox.question(self, title, message, QMessageBox.Cancel, QMessageBox.Ok)

        confirmation_dialog_result = self._confirmation_dialog_results[uuid_]

        result = False
        if reply == QMessageBox.Ok:
            result = True
        confirmation_dialog_result['result'] = result
        confirmation_dialog_result['semaphore'].release()

    def on_file_dialog(self, dialog_type, directory, allow_multiple, save_filename, file_filter):
        if dialog_type == FileDialog.FOLDER:
            self._file_name = QFileDialog.getExistingDirectory(
                self,
                self.localization['linux.openFolder'],
                directory,
                options=QFileDialog.ShowDirsOnly,
            )
        elif dialog_type == FileDialog.OPEN:
            if allow_multiple:
                self._file_name = QFileDialog.getOpenFileNames(
                    self, self.localization['linux.openFiles'], directory, file_filter
                )
            else:
                self._file_name = QFileDialog.getOpenFileName(
                    self, self.localization['linux.openFile'], directory, file_filter
                )
        elif dialog_type == FileDialog.SAVE:
            if directory:
                save_filename = os.path.join(str(directory), str(save_filename))

            self._file_name = QFileDialog.getSaveFileName(
                self, self.localization['global.saveFile'], save_filename
            )

        self._file_name_semaphore.release()

    def on_cookie_added(self, cookie):
        raw = str(cookie.toRawForm(), 'utf-8')
        cookie = create_cookie(raw)

        if raw not in self.cookies:
            self.cookies[raw] = cookie

    def on_cookie_removed(self, cookie):
        raw = str(cookie.toRawForm(), 'utf-8')

        if raw in self.cookies:
            del self.cookies[raw]

    def on_current_url(self):
        url = BrowserView._convert_string(self.webview.url().toString())
        self._current_url = None if url == '' or url.startswith('data:text/html') else url
        self._current_url_semaphore.release()

    def on_load_url(self, url):
        self.webview.setUrl(QtCore.QUrl(url))

    def on_load_html(self, content, base_uri):
        self.webview.setHtml(content, QtCore.QUrl(base_uri))

    def on_set_on_top(self, top):
        flags = self.windowFlags()
        if top:
            self.setWindowFlags(flags | QtCore.Qt.WindowStaysOnTopHint)
        else:
            self.setWindowFlags(flags & ~QtCore.Qt.WindowStaysOnTopHint)

        self.show()

    def showEvent(self, event):
        self.pywebview_window.events.shown.set()

    def closeEvent(self, event):
        should_cancel = self.pywebview_window.events.closing.set()

        if should_cancel:
            event.ignore()
            return

        if self.pywebview_window.confirm_close:
            reply = QMessageBox.question(
                self,
                self.title,
                self.localization['global.quitConfirmation'],
                QMessageBox.Yes,
                QMessageBox.No,
            )

            if reply == QMessageBox.No:
                event.ignore()
                return

        event.accept()

        del BrowserView.instances[self.uid]
        self.close()

        if self.pywebview_window in windows:
            windows.remove(self.pywebview_window)

        self.pywebview_window.events.closed.set()

        if self.webview.page():
            self.webview.page().deleteLater()

        if len(BrowserView.instances) == 0:
            self.hide()
            _app.exit()

    def changeEvent(self, e):
        if e.type() != QtCore.QEvent.WindowStateChange:
            return

        if self.windowState() == QtCore.Qt.WindowMinimized:
            self.pywebview_window.events.minimized.set()

        if self.windowState() == QtCore.Qt.WindowMaximized:
            self.pywebview_window.events.maximized.set()

        if self.windowState() == QtCore.Qt.WindowNoState and e.oldState() in (
            QtCore.Qt.WindowMinimized,
            QtCore.Qt.WindowMaximized,
        ):
            self.pywebview_window.events.restored.set()

    def resizeEvent(self, e):
        if (
            self.pywebview_window.initial_width != self.width()
            or self.pywebview_window.initial_height != self.height()
        ):
            self.pywebview_window.events.resized.set(self.width(), self.height())

    def eventFilter(self, object, event):
        if event.type() == QtCore.QEvent.Move:
            self.pywebview_window.events.moved.set(self.x(), self.y())

        return super().eventFilter(object, event)

    def on_show_window(self):
        self.show()

    def on_hide_window(self):
        self.hide()

    def on_destroy_window(self):
        self.close()

    def on_fullscreen(self):
        if self.is_fullscreen:
            self.showNormal()
        else:
            self.showFullScreen()

        self.is_fullscreen = not self.is_fullscreen

    def on_window_size(self, width, height, fix_point):
        geo = self.geometry()

        if fix_point & FixPoint.EAST:
            # Keep the right of the window in the same place
            geo.setX(geo.x() + geo.width() - width)

        if fix_point & FixPoint.SOUTH:
            # Keep the top of the window in the same place
            geo.setY(geo.y() + geo.height() - height)

        geo.setWidth(width)
        geo.setHeight(height)
        self.setGeometry(geo)

        if not self.pywebview_window.resizable:
            self.setFixedSize(width, height)

    def on_window_move(self, x, y):
        self.move(x, y)

    def on_window_maximize(self):
        self.setWindowState(QtCore.Qt.WindowMaximized)

    def on_window_minimize(self):
        self.setWindowState(QtCore.Qt.WindowMinimized)

    def on_window_restore(self):
        self.setWindowState(QtCore.Qt.WindowNoState)
        self.raise_()
        self.activateWindow()

    def on_evaluate_js(self, script, uuid):
        def return_result(result_):
            result = BrowserView._convert_string(result_)
            uuid_ = BrowserView._convert_string(uuid)

            js_result = self._js_results[uuid_]

            if js_result['parse_json'] and result:
                try:
                    js_result['result'] = json.loads(result)
                except Exception:
                    logger.exception('Failed to parse JSON: %s', result)
                    js_result['result'] = result
            else:
                js_result['result'] = result

            js_result['semaphore'].release()

        try:  # < Qt5.6
            if _qt6:
                self.webview.page().runJavaScript(script, 0, return_result)
            else:
                self.webview.page().runJavaScript(script, return_result)
        except TypeError:
            self.webview.page().runJavaScript(script)  # PySide2 & PySide6
        except AttributeError:
            result = self.webview.page().mainFrame().evaluateJavaScript(script)
            return_result(result)
        except Exception as e:
            logger.exception(e)

    def on_load_finished(self):
        if self.uid == 'web_inspector':
            return

        self._set_js_api()

        if _state['debug'] and settings['OPEN_DEVTOOLS_IN_DEBUG']:
            self.webview.show_inspector()

    def on_download_requested(self, download):
        old_path = download.url().path()
        suffix = QtCore.QFileInfo(old_path).suffix()
        path, _ = QFileDialog.getSaveFileName(
            self, self.localization['global.saveFile'], old_path, '*.' + suffix
        )
        if path:
            download.setPath(path)
            download.accept()

    def set_title(self, title):
        self.set_title_trigger.emit(title)

    def get_cookies(self):
        return list(self.cookies.values())

    def clear_cookies(self):
        self.cookies = {}
        self.profile.cookieStore().deleteAllCookies()

    def get_current_url(self):
        self.current_url_trigger.emit()
        self._current_url_semaphore.acquire()

        return self._current_url

    def load_url(self, url):
        self.load_url_trigger.emit(url)

    def load_html(self, content, base_uri):
        self.html_trigger.emit(content, base_uri)

    def create_confirmation_dialog(self, title, message):
        result_semaphore = Semaphore(0)
        unique_id = uuid1().hex
        self._confirmation_dialog_results[unique_id] = {
            'semaphore': result_semaphore,
            'result': None,
        }

        self.confirmation_dialog_trigger.emit(title, message, unique_id)
        result_semaphore.acquire()

        result = self._confirmation_dialog_results[unique_id]['result']
        del self._confirmation_dialog_results[unique_id]

        return result

    def create_file_dialog(
        self, dialog_type, directory, allow_multiple, save_filename, file_filter
    ):
        self.file_dialog_trigger.emit(
            dialog_type, directory, allow_multiple, save_filename, file_filter
        )
        self._file_name_semaphore.acquire()

        if dialog_type == FileDialog.FOLDER:
            file_names = (self._file_name,)
        elif dialog_type == FileDialog.SAVE or not allow_multiple:
            file_names = (self._file_name[0],)
        else:
            file_names = tuple(self._file_name[0])

        # Check if we got an empty tuple, or a tuple with empty string
        if len(file_names) == 0 or len(file_names[0]) == 0:
            return None
        else:
            return file_names

    def hide_(self):
        self.hide_trigger.emit()

    def show_(self):
        self.show_trigger.emit()

    def destroy_(self):
        self.destroy_trigger.emit()

    def toggle_fullscreen(self):
        self.fullscreen_trigger.emit()

    def resize_(self, width, height, fix_point):
        self.window_size_trigger.emit(width, height, fix_point)

    def move_window(self, x, y):
        self.window_move_trigger.emit(x, y)

    def maximize(self):
        self.window_maximize_trigger.emit()

    def minimize(self):
        self.window_minimize_trigger.emit()

    def restore(self):
        self.window_restore_trigger.emit()

    def set_on_top(self, top):
        self.on_top_trigger.emit(top)

    def evaluate_js(self, script, parse_json):
        result_semaphore = Semaphore(0)
        unique_id = uuid1().hex
        self._js_results[unique_id] = {
            'semaphore': result_semaphore,
            'result': '',
            'parse_json': parse_json,
        }

        self.evaluate_js_trigger.emit(script, unique_id)
        result_semaphore.acquire()

        result = deepcopy(self._js_results[unique_id]['result'])
        del self._js_results[unique_id]

        return result

    def _set_js_api(self):
        def _register_window_object():
            frame.addToJavaScriptWindowObject('external', self.js_bridge)

        if is_webengine:
            qwebchannel_js = QtCore.QFile('://qtwebchannel/qwebchannel.js')
            if qwebchannel_js.open(QtCore.QFile.ReadOnly):
                source = bytes(qwebchannel_js.readAll()).decode('utf-8')
                self.webview.page().runJavaScript(source)
                self.channel.registerObject('external', self.js_bridge)
                qwebchannel_js.close()
        else:
            frame = self.webview.page().mainFrame()
            _register_window_object()

        inject_pywebview(renderer, self.js_bridge.window)

    @staticmethod
    def _convert_string(result):
        if isinstance(result, QJsonValue):
            return None if result.isNull() else result.toString()
        else:
            return result

    @staticmethod
    def _get_debug_port():
        """
        Check if default debug port 8228 is available,
        increment it by 1 until a port is available.
        :return: port: str
        """
        port_available = False
        port = 8228

        while not port_available:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.bind(('localhost', port))
                port_available = True
            except Exception as e:
                port_available = False
                logger.warning(f'Port {port} is in use: {e}')
                port += 1
            finally:
                sock.close()

        return str(port)

    @staticmethod
    # Receive func from subthread and execute it on the main thread
    def on_create_window(func):
        func()


def setup_app():
    global _app
    if settings['IGNORE_SSL_ERRORS']:
        environ_append('QTWEBENGINE_CHROMIUM_FLAGS', '--ignore-certificate-errors')
    if settings['REMOTE_DEBUGGING_PORT']:
        environ_append(
            'QTWEBENGINE_CHROMIUM_FLAGS',
            f'--remote-debugging-port={settings["REMOTE_DEBUGGING_PORT"]}',
        )
    _app = QApplication.instance() or QApplication(sys.argv)


def create_window(window):
    def _create():
        browser = BrowserView(window)
        browser.installEventFilter(browser)

        if window.menu or _app_menu:
            menu = window.menu or _state['menu']
            window_menubar = browser.menuBar()
            create_menu(menu, window_menubar)

        _main_window_created.set()

        if window.maximized:
            browser.showMaximized()
        elif window.minimized:
            # showMinimized does not work on start without showNormal first
            # looks like a bug in QT
            browser.showNormal()
            browser.showMinimized()
        elif not window.hidden:
            browser.show()
        elif window.hidden:
            browser.pywebview_window.events.shown.set()

    if window.uid == 'master':
        global _app, _app_menu
        if _state['menu']:
            _app_menu = QMenuBar()
            _app_menu.setNativeMenuBar(False)
            create_menu(_state['menu'], _app_menu)
        else:
            _app_menu = None

        _app = QApplication.instance() or QApplication(sys.argv)

        signal.signal(signal.SIGINT, _sigint_handler)

        # Create a timer to periodically allow the Python interpreter to run
        # This enables the signal handler to be called even when the Qt event loop is running
        # The timer just calls a no-op function to trigger Python execution
        timer = QtCore.QTimer()
        timer.start(500)
        timer.timeout.connect(lambda: None)

        _create()
        _app.exec_()
    else:
        _main_window_created.wait()
        i = list(BrowserView.instances.values())[0]  # arbitrary instance
        i.create_window_trigger.emit(_create)


def set_title(title, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.set_title(title)


def clear_cookies(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.clear_cookies()


def get_cookies(uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.get_cookies()


def get_current_url(uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.get_current_url()


def load_url(url, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.load_url(url)


def load_html(content, base_uri, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.load_html(content, base_uri)


def create_menu(app_menu_list, menubar):
    """
    Create the menu bar for the application for the provided QMenuBar object. Menu can be either a global
    application menu or a window-specific menu.

    Args:
        app_menu_list ([webview.menu.Menu])
        menubar (QMenuBar)
    """

    def run_action(func):
        Thread(target=func).start()

    def create_submenu(title, line_items, supermenu):
        m = supermenu.addMenu(title)
        BrowserView.global_menubar_other_objects.append(m)
        for menu_line_item in line_items:
            if isinstance(menu_line_item, MenuSeparator):
                m.addSeparator()
            elif isinstance(menu_line_item, MenuAction):
                new_action = QAction(menu_line_item.title)
                func = copy(menu_line_item.function)
                new_action.triggered.connect(partial(run_action, func))
                m.addAction(new_action)
                BrowserView.global_menubar_other_objects.append(new_action)
            elif isinstance(menu_line_item, Menu):
                create_submenu(menu_line_item.title, menu_line_item.items, m)

        return m

    for app_menu in app_menu_list:
        # Ignore '__app__' menus (macOS-only feature)
        if app_menu.title == '__app__':
            continue
        menu = create_submenu(app_menu.title, app_menu.items, menubar)
        menubar.addMenu(menu)


def get_active_window():
    active_window = None
    try:
        active_window = _app.activeWindow()
    except Exception as e:
        logger.warning(f'Failed to get active window: {e}')
        return None

    if active_window:
        return active_window.pywebview_window

    return None


def destroy_window(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.destroy_()


def hide(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.hide_()


def show(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.show_()


def maximize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.maximize()


def minimize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.minimize()


def restore(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.restore()


def toggle_fullscreen(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.toggle_fullscreen()


def set_on_top(uid, top):
    i = BrowserView.instances.get(uid)
    if i:
        i.set_on_top(top)


def resize(width, height, uid, fix_point):
    i = BrowserView.instances.get(uid)
    if i:
        i.resize_(width, height, fix_point)


def move(x, y, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.move_window(x, y)


def create_confirmation_dialog(title, message, uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.create_confirmation_dialog(title, message)


def create_file_dialog(dialog_type, directory, allow_multiple, save_filename, file_types, uid):
    # Create a file filter by parsing allowed file types
    file_types = [s.replace(';', ' ') for s in file_types]
    file_filter = ';;'.join(file_types)

    i = BrowserView.instances.get(uid)
    if i:
        return i.create_file_dialog(
            dialog_type, directory, allow_multiple, save_filename, file_filter
        )


def evaluate_js(script, uid, parse_json=True):
    i = BrowserView.instances.get(uid)
    if i:
        return i.evaluate_js(script, parse_json)


def get_position(uid):
    i = BrowserView.instances.get(uid)
    if i:
        position = i.geometry()
        return position.x(), position.y()
    else:
        return None, None


def get_size(uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.width(), i.height()
    else:
        return None, None


def get_screens():
    global _app
    _app = QApplication.instance() or QApplication(sys.argv)
    screens = [
        Screen(
            s.geometry().x(),
            s.geometry().y(),
            s.geometry().width(),
            s.geometry().height(),
            s,
            s.devicePixelRatio(),
        )
        for s in _app.screens()
    ]

    return screens


def add_tls_cert(certfile):
    config = QSslConfiguration.defaultConfiguration()
    certs = config.caCertificates()
    cert = QSslCertificate.fromPath(certfile)[0]
    certs.append(cert)
    config.setCaCertificates(certs)
    QSslConfiguration.setDefaultConfiguration(config)



"""
This package will help reduce overhead startup time of pywebview android application
on android. It will be responsible for defining android and api signature to reduce
the over head cost of pyjnius finding all signatures by itself.
"""

__all__ = [
    'AlertDialogBuilder',
    'DownloadManagerRequest',
    'Context',
    'Uri',
    'Environment',
    'KeyEvent',
    'View',
    'CookieManager',
    'PyJavascriptInterface',
    'PyWebChromeClient',
    'PyWebViewClient',
    'WebView',
]

from webview.platforms.android.jclass.app import AlertDialogBuilder, DownloadManagerRequest
from webview.platforms.android.jclass.content import Context
from webview.platforms.android.jclass.net import Uri
from webview.platforms.android.jclass.os import Environment
from webview.platforms.android.jclass.view import KeyEvent, View
from webview.platforms.android.jclass.webkit import (
    CookieManager,
    PyJavascriptInterface,
    PyWebChromeClient,
    PyWebViewClient,
    WebView,
)



from __future__ import annotations

import inspect
import logging
import threading
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import Self

if TYPE_CHECKING:
    from typing import type_check_only


logger = logging.getLogger('pywebview')


class EventContainer:
    _serializable = False

    if TYPE_CHECKING:

        @type_check_only
        def __getattr__(self, __name: str) -> Event: ...

        @type_check_only
        def __setattr__(self, __name: str, __value: Event) -> None: ...


class Event:
    def __init__(self, window, should_lock: bool = False) -> None:
        self._items: list[Callable[..., Any]] = []
        self._should_lock = should_lock
        self._event = threading.Event()
        self._window = window

    def set(self, *args: Any, **kwargs: Any) -> bool:
        def execute():
            for func in self._items:
                try:
                    if len(inspect.signature(func).parameters.values()) == 0:
                        value = func()
                    elif 'window' in inspect.signature(func).parameters:
                        value = func(self._window, *args, **kwargs)
                    else:
                        value = func(*args, **kwargs)
                    return_values.add(value)

                except Exception as e:
                    logger.exception(e)

        return_values: set[Any] = set()

        if len(self._items):
            if self._should_lock:
                execute()
            else:
                t = threading.Thread(target=execute)
                t.start()

        false_values = [v for v in return_values if v is False]
        self._event.set()

        return len(false_values) != 0

    def is_set(self) -> bool:
        return self._event.is_set()

    def wait(self, timeout: float | None = None) -> bool:
        return self._event.wait(timeout)

    def clear(self) -> None:
        return self._event.clear()

    def __add__(self, item: Callable[..., Any]) -> Self:
        self._items.append(item)
        return self

    def __sub__(self, item: Callable[..., Any]) -> Self:
        self._items.remove(item)
        return self

    def __iadd__(self, item: Callable[..., Any]) -> Self:
        self._items.append(item)
        return self

    def __isub__(self, item: Callable[..., Any]) -> Self:
        self._items.remove(item)
        return self

    def __len__(self) -> int:
        return len(self._items)



__all__ = ('Environment',)

from jnius import JavaClass, JavaStaticField, MetaJavaClass


class Environment(JavaClass, metaclass=MetaJavaClass):
    """
    Represents the Android environment class for interacting with the external storage and system paths.

    This class provides access to system-defined constants and methods for
    interacting with the device's file and directory management environment.
    Acts as a proxy to the Android's Java environment class. Use its attributes
    and methods to access critical file paths or configurations within the
    Android operating system.

    Attributes:
        __javaclass__ (str): Represents the Java class path being bridged.
        DIRECTORY_DOWNLOADS (JavaStaticField): Static field representing the
            path for storing user downloads, defined in the Android
            environment.
    """

    __javaclass__ = 'android/os/Environment'
    DIRECTORY_DOWNLOADS = JavaStaticField('Ljava/lang/String;')



__all__ = (
    'PyWebViewClient',
    'PyWebChromeClient',
    'PyJavascriptInterface',
    'CookieManager',
    'WebView',
)

from jnius import (
    JavaClass,
    JavaMethod,
    JavaMultipleMethod,
    JavaStaticField,
    JavaStaticMethod,
    MetaJavaClass,
)


class PyWebViewClient(JavaClass, metaclass=MetaJavaClass):
    """
    Represents a Java WebView client wrapper for interaction with Python.

    This class serves as a bridge between a Java WebView client in the Android
    platform and Python. It allows setting callbacks for events and intercepting
    requests, enabling Python code to handle Java WebView behaviors dynamically.

    Attributes:
        __javaclass__: str
            The fully qualified Java class name for the associated Java object.

    Methods:
        setCallback:
            Sets the callback wrapper to handle specific WebView events, along
            with an option to decide whether to enable or disable a specific
            callback mechanism.

        setRequestInterceptor:
            Sets a request interceptor to manage and modify WebView requests
            before they are processed.
    """

    __javaclass__ = 'com/pywebview/PyWebViewClient'
    setCallback = JavaMethod('(Lcom/pywebview/EventCallbackWrapper;Z)V')
    setRequestInterceptor = JavaMethod('(Lcom/pywebview/WebViewRequestInterceptor;)V')


class PyWebChromeClient(JavaClass, metaclass=MetaJavaClass):
    """
    Handles communication with the WebView's JavaScript environment.

    The PyWebChromeClient class acts as a bridge between JavaScript running in a WebView
    and the Python environment. By defining methods that handle JavaScript interactions,
    such as alert dialogs and confirmation dialogs, this class provides mechanisms
    to integrate JavaScript behaviors into Python-based applications.

    Attributes:
        __javaclass__: The fully qualified class name of the Java class that this Python
                       bridge wraps for interaction.

    Methods:
        setCallback: Sets the callback mechanism for handling JavaScript events. Takes an
                     EventCallbackWrapper as an argument.
        onJsAlert: Handles JavaScript alert dialogs presented within a WebView. It processes
                   the dialog content and determines whether the alert is handled in
                   the Python environment.
        onJsConfirm: Handles JavaScript confirmation dialogs within a WebView, allowing the
                     Python application to process the user's response to the dialog.
    """

    __javaclass__ = 'com/pywebview/PyWebChromeClient'
    setCallback = JavaMethod('(Lcom/pywebview/EventCallbackWrapper;)V')


class PyJavascriptInterface(JavaClass, metaclass=MetaJavaClass):
    """
    Interface that acts as a bridge between Python and the JavaScript context.

    This class provides methods to set a callback and to handle JavaScript
    calls by interfacing with the Java-based functionality. It enables
    interactions between Python code and JavaScript running on a web view or
    similar component by defining Java methods that can be invoked from
    JavaScript.

    Attributes:
        __javaclass__ (str): Represents the Java class path of the associated
            Java implementation.

    Methods:
        setCallback: Registers a JavaScript API callback wrapper to handle
            responses from JavaScript to Python.
        call: Executes a specified JavaScript function with provided context
            and arguments.
    """

    __javaclass__ = 'com/pywebview/PyJavascriptInterface'
    setCallback = JavaMethod('(Lcom/pywebview/JsApiCallbackWrapper;)V')
    call = JavaMethod('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V')


class CookieManager(JavaClass, metaclass=MetaJavaClass):
    """
    Manages browser cookies for the WebView component and provides methods for
    cookie handling.

    The CookieManager class offers a variety of methods to control HTTP cookie
    management within a WebView environment. Using this class, you can query,
    set, accept, and delete cookies, as well as enable or disable the acceptance
    of third-party or file-scheme cookies. Cookie management features are essential
    for maintaining session persistence and enforcing privacy policies in web views
    embedded within Android applications.

    Attributes:
        __javaclass__ (str): Specifies the underlying Java class representation
            of the `android.webkit.CookieManager`.

    Methods:
        getInstance(): Retrieves the singleton instance of the CookieManager.

        setAcceptCookie(accept: bool): Sets whether the WebView should accept
            cookies globally.

        acceptCookie(accept: bool): Checks whether the web environment accepts cookies.

        setAcceptThirdPartyCookies(view, accept: bool): Determines whether the
            specified WebView should accept third-party cookies.

        acceptThirdPartyCookies(view): Indicates whether the given WebView accepts
            third-party cookies.

        setCookie(url: str, value: str, callback: Optional = None): Sets a cookie
            for the specified URL, optionally performing an action asynchronously
            upon completion.

        getCookie(url: str): Retrieves the cookie(s) associated with the specified
            URL.

        removeSessionCookie(): Removes all session cookies synchronously.

        removeSessionCookies(callback): Removes all session cookies asynchronously,
            triggering the provided ValueCallback upon completion.

        removeAllCookie(): Deletes all cookies synchronously.

        removeAllCookies(callback): Deletes all cookies asynchronously, using the
            provided ValueCallback to notify when the task is completed.

        hasCookies(): Indicates whether any cookies are stored.

        removeExpiredCookie(): Removes cookies that are no longer valid based on
            their expiration attributes.

        flush(): Forces synchronous writing of the cookie state to storage.

        allowFileSchemeCookies(): Checks if cookies for file-scheme URLs are
            accepted.

        setAcceptFileSchemeCookies(accept: bool): Enables or disables the
            acceptance of cookies for file-scheme URLs.
    """

    __javaclass__ = 'android/webkit/CookieManager'

    getInstance = JavaStaticMethod('()Landroid/webkit/CookieManager;')
    setAcceptCookie = JavaMethod('(Z)V')
    acceptCookie = JavaMethod('()V')
    setAcceptThirdPartyCookies = JavaMethod('(Landroid/webkit/WebView;Z)V')
    acceptThirdPartyCookies = JavaMethod('(Landroid/webkit/WebView;)Z')
    setCookie = JavaMultipleMethod(
        [
            ('(Ljava/lang/String;Ljava/lang/String;)V', False, False),
            ('(Ljava/lang/String;Ljava/lang/String;Landroid/webkit/ValueCallback;)V', False, False),
        ]
    )
    getCookie = JavaMethod('(Ljava/lang/String;)Ljava/lang/String;')
    removeSessionCookie = JavaMethod('()V')
    removeSessionCookies = JavaMethod('(Landroid/webkit/ValueCallback;)V')
    removeAllCookie = JavaMethod('()V')
    removeAllCookies = JavaMethod('(Landroid/webkit/ValueCallback;)V')
    hasCookies = JavaMethod('()Z')
    removeExpiredCookie = JavaMethod('()V')
    flush = JavaMethod('()V')
    allowFileSchemeCookies = JavaStaticMethod('()Z')
    setAcceptFileSchemeCookies = JavaStaticMethod('(Z)V')


class WebView(JavaClass, metaclass=MetaJavaClass):
    __javaclass__ = 'android/webkit/WebView'
    __javaconstructor__ = [
        ('(Landroid/content/Context;)V', False),
        ('(Landroid/content/Context;Landroid/util/AttributeSet;)V', False),
        ('(Landroid/content/Context;Landroid/util/AttributeSet;I)V', False),
        ('(Landroid/content/Context;Landroid/util/AttributeSet;II)V', False),
        ('(Landroid/content/Context;Landroid/util/AttributeSet;IZ)V', False),
    ]

    RENDERER_PRIORITY_BOUND = JavaStaticField('I')
    RENDERER_PRIORITY_IMPORTANT = JavaStaticField('I')
    RENDERER_PRIORITY_WAIVED = JavaStaticField('I')
    SCHEME_GEO = JavaStaticField('Ljava/lang/String;')
    SCHEME_MAILTO = JavaStaticField('Ljava/lang/String;')
    SCHEME_TEL = JavaStaticField('Ljava/lang/String;')

    setHorizontalScrollbarOverlay = JavaMethod('(Z)V')
    setVerticalScrollbarOverlay = JavaMethod('(Z)V')
    overlayHorizontalScrollbar = JavaMethod('()Z')
    overlayVerticalScrollbar = JavaMethod('()Z')
    getCertificate = JavaMethod('()Landroid/net/http/SslCertificate;')
    setCertificate = JavaMethod('(Landroid/net/http/SslCertificate;)V')
    savePassword = JavaMethod('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V')
    setHttpAuthUsernamePassword = JavaMethod(
        '(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V'
    )
    getHttpAuthUsernamePassword = JavaMethod(
        '(Ljava/lang/String;Ljava/lang/String;)[Ljava/lang/String;'
    )
    destroy = JavaMethod('()V')
    setNetworkAvailable = JavaMethod('(Z)V')
    saveState = JavaMethod('(Landroid/os/Bundle;)Landroid/webkit/WebBackForwardList;')
    restoreState = JavaMethod('(Landroid/os/Bundle;)Landroid/webkit/WebBackForwardList;')

    loadUrl = JavaMultipleMethod(
        [
            ('(Ljava/lang/String;)V', False, False),
            ('(Ljava/lang/String;Ljava/util/Map;)V', False, False),
        ]
    )

    postUrl = JavaMethod('(Ljava/lang/String;[B)V')
    loadData = JavaMethod('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V')
    loadDataWithBaseURL = JavaMethod(
        '(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V'
    )
    evaluateJavascript = JavaMethod('(Ljava/lang/String;Landroid/webkit/ValueCallback;)V')
    saveWebArchive = JavaMultipleMethod(
        [
            ('(Ljava/lang/String;)V', False, False),
            ('(Ljava/lang/String;ZLandroid/webkit/ValueCallback;)V', False, False),
        ]
    )
    stopLoading = JavaMethod('()V')
    reload = JavaMethod('()V')
    canGoBack = JavaMethod('()Z')
    goBack = JavaMethod('()V')
    canGoForward = JavaMethod('()Z')
    goForward = JavaMethod('()V')
    canGoBackOrForward = JavaMethod('(I)Z')
    goBackOrForward = JavaMethod('(I)V')
    isPrivateBrowsingEnabled = JavaMethod('()Z')
    pageUp = JavaMethod('(Z)Z')
    pageDown = JavaMethod('(Z)Z')
    postVisualStateCallback = JavaMethod('(JLandroid/webkit/WebView$VisualStateCallback;)V')
    clearView = JavaMethod('()V')
    capturePicture = JavaMethod('()Landroid/graphics/Picture;')
    createPrintDocumentAdapter = JavaMultipleMethod(
        [
            ('()Landroid/print/PrintDocumentAdapter;', False, False),
            ('(Ljava/lang/String;)Landroid/print/PrintDocumentAdapter;', False, False),
        ]
    )
    getScale = JavaMethod('()F')
    setInitialScale = JavaMethod('(I)V')
    invokeZoomPicker = JavaMethod('()V')
    getHitTestResult = JavaMethod('()Landroid/webkit/WebView$HitTestResult;')
    requestFocusNodeHref = JavaMethod('(Landroid/os/Message;)V')
    requestImageRef = JavaMethod('(Landroid/os/Message;)V')
    getUrl = JavaMethod('()Ljava/lang/String;')
    getOriginalUrl = JavaMethod('()Ljava/lang/String;')
    getTitle = JavaMethod('()Ljava/lang/String;')
    getFavicon = JavaMethod('()Landroid/graphics/Bitmap;')
    getProgress = JavaMethod('()I')
    getContentHeight = JavaMethod('()I')
    getHeight = JavaMethod('()I')
    getWidth = JavaMethod('()I')
    pauseTimers = JavaMethod('()V')
    resumeTimers = JavaMethod('()V')
    onPause = JavaMethod('()V')
    onResume = JavaMethod('()V')
    freeMemory = JavaMethod('()V')
    clearCache = JavaMethod('(Z)V')
    clearFormData = JavaMethod('()V')
    clearHistory = JavaMethod('()V')
    clearSslPreferences = JavaMethod('()V')
    clearClientCertPreferences = JavaStaticMethod('(Ljava/lang/Runnable;)V')
    startSafeBrowsing = JavaStaticMethod(
        '(Landroid/content/Context;Landroid/webkit/ValueCallback;)V'
    )
    setSafeBrowsingWhitelist = JavaStaticMethod('(Ljava/util/List;Landroid/webkit/ValueCallback;)V')
    getSafeBrowsingPrivacyPolicyUrl = JavaStaticMethod('()Landroid/net/Uri;')
    copyBackForwardList = JavaMethod('()Landroid/webkit/WebBackForwardList;')
    setFindListener = JavaMethod('(Landroid/webkit/WebView$FindListener;)V')
    findNext = JavaMethod('(Z)V')
    findAll = JavaMethod('(Ljava/lang/String;)I')
    findAllAsync = JavaMethod('(Ljava/lang/String;)V')
    showFindDialog = JavaMethod('(Ljava/lang/String;Z)Z')
    findAddress = JavaStaticMethod('(Ljava/lang/String;)Ljava/lang/String;')
    enableSlowWholeDocumentDraw = JavaStaticMethod('()V')

    clearMatches = JavaMethod('()V')
    documentHasImages = JavaMethod('(Landroid/os/Message;)V')
    setWebViewClient = JavaMethod('(Landroid/webkit/WebViewClient;)V')
    getWebViewClient = JavaMethod('()Landroid/webkit/WebViewClient;')
    getWebViewRenderProcess = JavaMethod('()Landroid/webkit/WebViewRenderProcess;')
    setWebViewRenderProcessClient_with_executor = JavaMethod(
        'Ljava/util/concurrent/Executor;Landroid/webkit/WebViewRenderProcessClient;V'
    )
    setWebViewRenderProcessClient = JavaMultipleMethod(
        [
            (
                '(Ljava/util/concurrent/Executor;Landroid/webkit/WebViewRenderProcessClient;)V',
                False,
                False,
            ),
            ('Landroid/webkit/WebViewRenderProcessClient;V', False, False),
        ]
    )
    getWebViewRenderProcessClient = JavaMethod('()Landroid/webkit/WebViewRenderProcessClient;')
    setDownloadListener = JavaMethod('(Landroid/webkit/DownloadListener;)V')
    setWebChromeClient = JavaMethod('(Landroid/webkit/WebChromeClient;)V')
    getWebChromeClient = JavaMethod('()Landroid/webkit/WebChromeClient;')
    setPictureListener = JavaMethod('(Landroid/webkit/WebView$PictureListener;)V')
    setOnKeyListener = JavaMethod('(Landroid/view/View$OnKeyListener;)V')
    addJavascriptInterface = JavaMethod('(Ljava/lang/Object;Ljava/lang/String;)V')
    removeJavascriptInterface = JavaMethod('(Ljava/lang/String;)V')
    createWebMessageChannel = JavaMethod('()[Landroid/webkit/WebMessagePort;')
    postWebMessage = JavaMethod('(Landroid/webkit/WebMessage;Landroid/net/Uri;)V')
    getSettings = JavaMethod('()Landroid/webkit/WebSettings;')
    setWebContentsDebuggingEnabled = JavaStaticMethod('(Z)V')
    setDataDirectorySuffix = JavaStaticMethod('(Ljava/lang/String;)V')
    disableWebView = JavaStaticMethod('()V')
    onChildViewAdded = JavaMethod('(Landroid/view/View;Landroid/view/View;)V')
    onChildViewRemoved = JavaMethod('(Landroid/view/View;Landroid/view/View;)V')
    onGlobalFocusChanged = JavaMethod('(Landroid/view/View;Landroid/view/View;)V')
    setMapTrackballToArrowKeys = JavaMethod('(Z)V')
    flingScroll = JavaMethod('(II)V')
    canZoomIn = JavaMethod('()Z')
    canZoomOut = JavaMethod('()Z')
    zoomBy = JavaMethod('(F)V')
    zoomIn = JavaMethod('()Z')
    zoomOut = JavaMethod('()Z')
    setRendererPriorityPolicy = JavaMethod('(IZ)V')
    getRendererRequestedPriority = JavaMethod('()I')
    getRendererPriorityWaivedWhenNotVisible = JavaMethod('()Z')
    setTextClassifier = JavaMethod('(Landroid/view/textclassifier/TextClassifier;)V')
    getTextClassifier = JavaMethod('()Landroid/view/textclassifier/TextClassifier;')
    getWebViewClassLoader = JavaStaticMethod('()Ljava/lang/ClassLoader;')
    getWebViewLooper = JavaMethod('()Landroid/os/Looper;')
    onAttachedToWindow = JavaMethod('()V')
    setLayoutParams = JavaMethod('(Landroid/view/ViewGroup$LayoutParams;)V')
    setOverScrollMode = JavaMethod('(I)V')
    setScrollBarStyle = JavaMethod('(I)V')
    computeScroll = JavaMethod('()V')
    onHoverEvent = JavaMethod('(Landroid/view/MotionEvent;)Z')
    onTouchEvent = JavaMethod('(Landroid/view/MotionEvent;)Z')
    onGenericMotionEvent = JavaMethod('(Landroid/view/MotionEvent;)Z')
    onTrackballEvent = JavaMethod('(Landroid/view/MotionEvent;)Z')
    onKeyDown = JavaMethod('(ILandroid/view/KeyEvent;)Z')
    onKeyUp = JavaMethod('(ILandroid/view/KeyEvent;)Z')
    onKeyMultiple = JavaMethod('(IILandroid/view/KeyEvent;)Z')
    getAccessibilityNodeProvider = JavaMethod(
        '()Landroid/view/accessibility/AccessibilityNodeProvider;'
    )
    shouldDelayChildPressedState = JavaMethod('()Z')
    getAccessibilityClassName = JavaMethod('()Ljava/lang/CharSequence;')
    onProvideVirtualStructure = JavaMethod('(Landroid/view/ViewStructure;)V')
    onProvideAutofillVirtualStructure = JavaMethod('(Landroid/view/ViewStructure;I)V')
    onProvideContentCaptureStructure = JavaMethod('(Landroid/view/ViewStructure;I)V')
    autofill = JavaMethod('(Landroid/util/SparseArray;)V')
    isVisibleToUserForAutofill = JavaMethod('(I)Z')
    onCreateVirtualViewTranslationRequests = JavaMethod('([J[ILjava/util/function/Consumer;)V')
    dispatchCreateViewTranslationRequest = JavaMethod(
        '(Ljava/util/Map;[ILandroid/view/translation/TranslationCapability;Ljava/util/List;)V'
    )
    onVirtualViewTranslationResponses = JavaMethod('(Landroid/util/LongSparseArray;)V')
    onOverScrolled = JavaMethod('(IIZZ)V')
    onWindowVisibilityChanged = JavaMethod('(I)V')
    onDraw = JavaMethod('(Landroid/graphics/Canvas;)V')
    performLongClick = JavaMethod('()Z')
    onConfigurationChanged = JavaMethod('(Landroid/content/res/Configuration;)V')
    onCreateInputConnection = JavaMethod(
        '(Landroid/view/inputmethod/EditorInfo;)Landroid/view/inputmethod/InputConnection;'
    )
    onDragEvent = JavaMethod('(Landroid/view/DragEvent;)Z')
    onVisibilityChanged = JavaMethod('(Landroid/view/View;I)V')
    onWindowFocusChanged = JavaMethod('(Z)V')
    onFocusChanged = JavaMethod('(ZILandroid/graphics/Rect;)V')
    onSizeChanged = JavaMethod('(IIII)V')
    onScrollChanged = JavaMethod('(IIII)V')
    dispatchKeyEvent = JavaMethod('(Landroid/view/KeyEvent;)Z')
    requestFocus = JavaMethod('(ILandroid/graphics/Rect;)Z')
    onMeasure = JavaMethod('(II)V')
    requestChildRectangleOnScreen = JavaMethod('(Landroid/view/View;Landroid/graphics/Rect;Z)Z')
    setBackgroundColor = JavaMethod('(I)V')
    setLayerType = JavaMethod('(ILandroid/graphics/Paint;)V')
    dispatchDraw = JavaMethod('(Landroid/graphics/Canvas;)V')
    onStartTemporaryDetach = JavaMethod('()V')
    onFinishTemporaryDetach = JavaMethod('()V')
    getHandler = JavaMethod('()Landroid/os/Handler;')
    findFocus = JavaMethod('()Landroid/view/View;')
    getCurrentWebViewPackage = JavaStaticMethod('()Landroid/content/pm/PackageInfo;')
    onCheckIsTextEditor = JavaMethod('()Z')
    onApplyWindowInsets = JavaMethod('(Landroid/view/WindowInsets;)Landroid/view/WindowInsets;')
    onResolvePointerIcon = JavaMethod('(Landroid/view/MotionEvent;I)Landroid/view/PointerIcon;')



class EventDispatcher:
    __event_stack = {}

    def register_event_type(self, event_type):
        """Register an event type with the dispatcher.
        Registering event types allows the dispatcher to validate event handler
        names as they are attached and to search attached objects for suitable
        handlers. Each event type declaration must:
            1. start with the prefix `on_`.
            2. have a default handler in the class.
        """

        if event_type[:3] != 'on_':
            raise Exception('A new event must start with "on_"')

        # Ensure that the user has at least declared the default handler
        if not hasattr(self, event_type):
            raise Exception(f'Missing default handler {event_type} in {self.__class__.__name__}')

        # Add the event type to the stack
        if event_type not in self.__event_stack:
            self.__event_stack[event_type] = []

    def unregister_event_type(self, event_type):
        pass

    def bind(self, **kwargs):
        for key, value in kwargs.items():
            assert callable(value), f'{value!r} is not callable'
            if key[:3] == 'on_':
                observers: list = self.__event_stack.get(key)
                if observers is None:
                    continue
                observers.append(value)

    def unbind(self, **kwargs):
        """Unbind event from callback functions with similar usage as"""

        for key, value in kwargs.items():
            if key[:3] == 'on_':
                observers: list = self.__event_stack.get(key)
                if observers is None:
                    continue
                observers.remove(value)

    def is_event_type(self, event_type):
        """Return True if the event_type is already registered."""
        return event_type in self.__event_stack

    def dispatch(self, event_type, *args, **kwargs):
        """Dispatch an event across all the handlers added in bind/fbind().
        As soon as a handler returns True, the dispatching stops.
        The function collects all the positional and keyword arguments and
        passes them on to the handlers.
        .. note::
           The handlers are called in reverse order than they were registered
           with :meth:`bind`.
        :Parameters:
           `event_type`: str
               the event name to dispatch.
        .. versionchanged:: 1.9.0
           Keyword arguments collection and forwarding was added. Before, only
           positional arguments would be collected and forwarded.
        """
        observers: list = self.__event_stack.get(event_type)

        if observers:
            observers.reverse()
            for callbacks in observers:
                callbacks(*args, **kwargs)

        handler = getattr(self, event_type)
        return handler(*args, **kwargs)



"""
pywebview is a lightweight cross-platform wrapper around a webview component that allows to display HTML content in its
own dedicated window. Works on Windows, OS X and Linux and compatible with Python 2 and 3.

(C) 2014-2019 Roman Sirokov and contributors
Licensed under BSD license

http://github.com/r0x0r/pywebview/
"""

from __future__ import annotations

import datetime
import enum
import logging
import os
import re
import tempfile
import threading
from collections.abc import Iterable, Mapping
from typing import Any, Callable
from uuid import uuid4

from proxy_tools import module_property

import webview.http as http
from webview.errors import JavascriptException, WebViewException
from webview.event import Event
from webview.guilib import GUIType, initialize
from webview.localization import original_localization
from webview.menu import Menu
from webview.screen import Screen
from webview.util import _TOKEN, ImmutableDict, abspath, is_app, is_local_url
from webview.window import Window

__all__ = (
    # Stuff that's here
    'active_window',
    'start',
    'create_window',
    'token',
    'renderer',
    'screens',
    'settings',
    # From event
    'Event',
    # from util    '
    'JavascriptException',
    'WebViewException',
    # from screen
    'Screen',
    # from window
    'Window',
)


def _setup_logger():
    """Setup logger with console handler and appropriate log level."""

    logger = logging.getLogger('pywebview')

    # Avoid duplicate setup
    if logger.handlers:
        return logger

    # Create and configure handler
    handler = logging.StreamHandler()
    formatter = logging.Formatter('[pywebview] %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Set log level from environment variable with validation
    log_level_name = os.environ.get('PYWEBVIEW_LOG', 'INFO').upper()
    try:
        log_level = getattr(logging, log_level_name)
        logger.setLevel(log_level)
    except AttributeError:
        # Fallback to INFO if invalid level specified
        logger.setLevel(logging.INFO)
        logger.warning(f"Invalid log level '{log_level_name}', using INFO instead")

    return logger


logger = _setup_logger()


@module_property
def OPEN_DIALOG():
    logger.warning(
        "OPEN_DIALOG is deprecated and will be removed in a future version. Use 'FileDialog.OPEN' instead."
    )
    return 10


@module_property
def FOLDER_DIALOG():
    logger.warning(
        "FOLDER_DIALOG is deprecated and will be removed in a future version. Use 'FileDialog.FOLDER' instead."
    )
    return 20


@module_property
def SAVE_DIALOG():
    logger.warning(
        "SAVE_DIALOG is deprecated and will be removed in a future version. Use 'FileDialog.SAVE' instead."
    )
    return 30


class FileDialog(enum.IntEnum):
    OPEN = 10
    FOLDER = 20
    SAVE = 30


settings = ImmutableDict(
    {
        'ALLOW_DOWNLOADS': False,
        'ALLOW_FILE_URLS': True,
        'DRAG_REGION_SELECTOR': '.pywebview-drag-region',
        'DRAG_REGION_DIRECT_TARGET_ONLY': False,
        'DEFAULT_HTTP_PORT': 42001,
        'OPEN_EXTERNAL_LINKS_IN_BROWSER': True,
        'OPEN_DEVTOOLS_IN_DEBUG': True,
        'REMOTE_DEBUGGING_PORT': None,
        'IGNORE_SSL_ERRORS': False,
        'SHOW_DEFAULT_MENUS': True,
        'WEBVIEW2_RUNTIME_PATH': None,
    }
)


_state = ImmutableDict(
    {
        'debug': False,
        'storage_path': None,
        'private_mode': True,
        'user_agent': None,
        'http_server': False,
        'ssl': False,
        'icon': None,
        'menu': None,
    }
)


@module_property
def DRAG_REGION_SELECTOR():
    logger.warning(
        'DRAG_REGION_SELECTOR is deprecated and will be removed in a future version. Use \'settings["DRAG_REGION_SELECTOR"]\' instead.'
    )
    return settings['DRAG_REGION_SELECTOR']


guilib = None

token = _TOKEN
windows: list[Window] = []
renderer: str | None = None


def start(
    func: Callable[..., None] | None = None,
    args: Iterable[Any] | None = None,
    localization: dict[str, str] = {},
    gui: GUIType | None = None,
    debug: bool = False,
    http_server: bool = False,
    http_port: int | None = None,
    user_agent: str | None = None,
    private_mode: bool = True,
    storage_path: str | None = None,
    menu: list[Menu] = [],
    server: type[http.ServerType] = http.BottleServer,
    server_args: dict[Any, Any] = {},
    ssl: bool = False,
    icon: str | None = None,
):
    """
    Start a GUI loop and display previously created windows. This function must
    be called from a main thread.

    :param func: Function to invoke upon starting the GUI loop.
    :param args: Function arguments. Can be either a single value or a tuple of
        values.
    :param localization: A dictionary with localized strings. Default strings
        and their keys are defined in localization.py.
    :param gui: Force a specific GUI. Allowed values are ``cef``, ``qt``,
        ``gtk``, ``mshtml`` or ``edgechromium`` depending on a platform.
    :param debug: Enable debug mode. Default is False.
    :param http_server: Enable built-in HTTP server. If enabled, local files
        will be served using a local HTTP server on a random port. For each
        window, a separate HTTP server is spawned. This option is ignored for
        non-local URLs.
    :param user_agent: Change user agent string.
    :param private_mode: Enable private mode. In private mode, cookies and local storage are not preserved.
           Default is True.
    :param storage_path: Custom location for cookies and other website data
    :param menu: List of menus to be included in the app menu
    :param server: Server class. Defaults to BottleServer
    :param server_args: Dictionary of arguments to pass through to the server instantiation
    :param ssl: Enable SSL for local HTTP server. Default is False.
    :param icon: Path to the icon file. Supported only on GTK/QT.
    """
    global guilib, renderer

    def _create_children(other_windows):
        if not windows[0].events.shown.wait(10):
            raise WebViewException('Main window failed to load')

        for window in other_windows:
            guilib.create_window(window)

    _state['debug'] = debug
    _state['user_agent'] = user_agent
    _state['http_server'] = http_server
    _state['private_mode'] = private_mode

    if icon:
        _state['icon'] = abspath(icon)

    if storage_path:
        __set_storage_path(storage_path)

    if debug and not os.environ.get('PYWEBVIEW_LOG'):
        logger.setLevel(logging.DEBUG)

    if (
        _state['storage_path']
        and _state['private_mode']
        and not os.path.exists(_state['storage_path'])
    ):
        os.makedirs(_state['storage_path'])

    original_localization.update(localization)

    if threading.current_thread().name != 'MainThread':
        raise WebViewException('pywebview must be run on a main thread.')

    if len(windows) == 0:
        raise WebViewException('You must create a window first before calling this function.')

    if menu:
        _state['menu'] = menu

    guilib = initialize(gui)
    renderer = guilib.renderer

    if ssl:
        if not server_args or 'keyfile' not in server_args or 'certfile' not in server_args:
            # generate SSL certs and tell the windows to use them
            keyfile, certfile = __generate_ssl_cert()
            server_args['keyfile'] = keyfile
            server_args['certfile'] = certfile
        else:
            keyfile = server_args['keyfile']
            certfile = server_args['certfile']
            if not os.path.exists(keyfile):
                raise WebViewException(f'The {keyfile} does not exist.')
            if not os.path.exists(certfile):
                raise WebViewException(f'The {certfile} does not exist.')

        _state['ssl'] = True
    else:
        keyfile, certfile = None, None
        server_args.pop('keyfile', None)
        server_args.pop('certfile', None)

    urls = [w.original_url for w in windows]
    has_local_urls = not not [w.original_url for w in windows if is_local_url(w.original_url)]
    # start the global server if it's not running and we need it
    if (http.global_server is None) and (http_server or has_local_urls):
        if not _state['private_mode'] and not http_port:
            http_port = settings['DEFAULT_HTTP_PORT']
        *_, server = http.start_global_server(
            http_port=http_port, urls=urls, server=server, **server_args
        )

    for window in windows:
        should_initialize = not window._initialize(guilib, server_args=server_args)
        if should_initialize:
            return

    if ssl:
        for window in windows:
            window.gui.add_tls_cert(certfile)

    if len(windows) > 1:
        thread = threading.Thread(target=_create_children, args=(windows[1:],))
        thread.start()

    if func:
        if args is not None:
            if not hasattr(args, '__iter__'):
                args = (args,)
            thread = threading.Thread(target=func, args=args)
        else:
            thread = threading.Thread(target=func)
        thread.start()

    guilib.create_window(windows[0])
    # keyfile is deleted by the ServerAdapter right after wrap_socket()
    if certfile:
        os.unlink(certfile)


def create_window(
    title: str,
    url: str | callable | None = None,
    html: str | None = None,
    js_api: Any = None,
    width: int = 800,
    height: int = 600,
    x: int | None = None,
    y: int | None = None,
    screen: Screen | None = None,
    resizable: bool = True,
    fullscreen: bool = False,
    min_size: tuple[int, int] = (200, 100),
    hidden: bool = False,
    frameless: bool = False,
    easy_drag: bool = True,
    shadow: bool = True,
    focus: bool = True,
    minimized: bool = False,
    maximized: bool = False,
    on_top: bool = False,
    confirm_close: bool = False,
    background_color: str = '#FFFFFF',
    transparent: bool = False,
    text_select: bool = False,
    zoomable: bool = False,
    draggable: bool = False,
    vibrancy: bool = False,
    menu: list[Menu] = [],
    localization: Mapping[str, str] | None = None,
    server: type[http.ServerType] = http.BottleServer,
    http_port: int | None = None,
    server_args: http.ServerArgs = {},
) -> Window | None:
    """
    Create a web view window using a native GUI. The execution blocks after this function is invoked, so other
    program logic must be executed in a separate thread.
    :param title: Window title
    :param url: URL or WSGI/ASGI app to load
    :param html: HTML content to load
    :param width: window width. Default is 800px
    :param height: window height. Default is 600px
    :param screen: Screen to display the window on.
    :param resizable: True if window can be resized, False otherwise. Default is True
    :param fullscreen: True if start in fullscreen mode. Default is False
    :param min_size: a (width, height) tuple that specifies a minimum window size. Default is 200x100
    :param hidden: Whether the window should be hidden.
    :param frameless: Whether the window should have a frame.
    :param easy_drag: Easy window drag mode when window is frameless.
    :param shadow: Whether the window should have a frame border (shadows and Windows rounded edges).
    :param focus: Whether to activate the window when user opens it. Window can be controlled with mouse but keyboard input will go to another (active) window and not this one.
    :param minimized: Display window minimized
    :param maximized: Display window maximized
    :param on_top: Keep window above other windows (required OS: Windows)
    :param confirm_close: Display a window close confirmation dialog. Default is False
    :param background_color: Background color as a hex string that is displayed before the content of webview is loaded. Default is white.
    :param text_select: Allow text selection on page. Default is False.
    :param transparent: Don't draw window background.
    :param menu: List of menus to be included in the window menu
    :param server: Server class. Defaults to BottleServer
    :param server_args: Dictionary of arguments to pass through to the server instantiation
    :return: window object or None if window initialization is cancelled in the window.events.initialized event
    """

    valid_color = r'^#(?:[0-9a-fA-F]{3}){1,2}$'
    if not re.match(valid_color, background_color):
        raise ValueError(f'{background_color} is not a valid hex triplet color')

    uid = 'master' if len(windows) == 0 else 'child_' + uuid4().hex[:8]

    window = Window(
        uid,
        title,
        url,
        html,
        width,
        height,
        x,
        y,
        resizable,
        fullscreen,
        min_size,
        hidden,
        frameless,
        easy_drag,
        shadow,
        focus,
        minimized,
        maximized,
        on_top,
        confirm_close,
        background_color,
        js_api,
        text_select,
        transparent,
        zoomable,
        draggable,
        vibrancy,
        menu,
        localization,
        server=server,
        http_port=http_port,
        server_args=server_args,
        screen=screen,
    )

    windows.append(window)

    # This immediately creates the window only if `start` has already been called
    if threading.current_thread().name != 'MainThread' and guilib:
        if is_app(url) or is_local_url(url) and not server.is_running:
            _, _, server = http.start_server([url], server=server, **server_args)
        else:
            server = None

        if not window._initialize(gui=guilib, server=server):
            return
        guilib.create_window(window)

    return window


def __generate_ssl_cert():
    try:
        # https://cryptography.io/en/latest/x509/tutorial/#creating-a-self-signed-certificate
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.x509.oid import NameOID
    except ImportError:
        raise WebViewException(
            'SSL support requires cryptography package. Please install it with "pip install cryptography" or "pip install pywebview[ssl]"'
        )

    with tempfile.NamedTemporaryFile(prefix='keyfile_', suffix='.pem', delete=False) as f:
        keyfile = f.name
        key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        key_pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),  # BestAvailableEncryption(b"passphrase"),
        )
        f.write(key_pem)

    with tempfile.NamedTemporaryFile(prefix='certfile_', suffix='.pem', delete=False) as f:
        certfile = f.name
        subject = issuer = x509.Name(
            [
                x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'California'),
                x509.NameAttribute(NameOID.LOCALITY_NAME, 'San Francisco'),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'pywebview'),
                x509.NameAttribute(NameOID.COMMON_NAME, '127.0.0.1'),
            ]
        )
        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
            .add_extension(
                x509.SubjectAlternativeName([x509.DNSName('localhost')]),
                critical=False,
            )
            .sign(key, hashes.SHA256(), backend=default_backend())
        )
        cert_pem = cert.public_bytes(serialization.Encoding.PEM)
        f.write(cert_pem)

    return keyfile, certfile


def __set_storage_path(storage_path):
    e = WebViewException(f'Storage path {storage_path} is not writable')

    if not os.path.exists(storage_path):
        try:
            os.makedirs(storage_path)
        except OSError:
            raise e
    if not os.access(storage_path, os.W_OK):
        raise e

    _state['storage_path'] = storage_path


def active_window() -> Window | None:
    """
    Get the active window

    :return: window object or None
    """
    if guilib:
        return guilib.get_active_window()
    return None


@module_property
def screens() -> list[Screen]:
    global renderer, guilib

    if not guilib:
        guilib = initialize()
        renderer = guilib.renderer

    screens = guilib.get_screens()
    return screens



import json
import logging
from http.cookies import SimpleCookie
from threading import Semaphore
from urllib.parse import urlparse

from android.activity import _activity as activity  # noqa
from android.runnable import run_on_ui_thread  # noqa
from jnius import autoclass, cast

from webview import _state, settings
from webview.models import Request, Response
from webview.platforms.android.app import App
from webview.platforms.android.jclass import (
    AlertDialogBuilder,
    Context,
    CookieManager,
    DownloadManagerRequest,
    Environment,
    KeyEvent,
    PyJavascriptInterface,
    PyWebChromeClient,
    PyWebViewClient,
    Uri,
    View,
    WebView,
)
from webview.platforms.android.jinterface import (
    DownloadListener,
    EventCallbackWrapper,
    JsApiCallbackWrapper,
    KeyListener,
    RequestInterceptor,
    ValueCallback,
)
from webview.util import create_cookie, inject_pywebview, js_bridge_call

logger = logging.getLogger('pywebview')

renderer = 'android-webkit'
app = None


class BrowserView:
    def __init__(self, window):
        self.pywebview_window = window
        self.webview = None
        self.dialog = None
        self.pywebview_window.native = self
        self.is_fullscreen = False
        self.create_webview()

    @run_on_ui_thread
    def create_webview(self):
        def js_api_handler(func, params, id):
            js_bridge_call(self.pywebview_window, func, json.loads(params), id)

        def chrome_callback(event, data):
            print(event, data)

        def webview_callback(event, data):
            if event == 'onPageFinished':

                @run_on_ui_thread
                def _handle_page_finished():
                    try:
                        inject_pywebview(renderer, self.pywebview_window)
                        # Get fresh CookieManager reference to avoid stale Java object references
                        cookie_manager = CookieManager.getInstance()
                        if not _state['private_mode']:
                            cookie_manager.setAcceptCookie(True)
                            cookie_manager.acceptCookie()
                            cookie_manager.flush()
                        else:
                            cookie_manager.setAcceptCookie(False)
                    except Exception as e:
                        logger.error(f'Error handling page finished: {e}')

                _handle_page_finished()

            elif event == 'onCookiesReceived':
                try:
                    cookie_data = json.loads(data) if data else {}
                    # url = cookie_data.get('url', '')
                    cookies = cookie_data.get('cookies', [])
                    for cookie_string in cookies:
                        cookie = SimpleCookie()
                        cookie.load(cookie_string)
                        app.view._cookies.append(cookie)

                except Exception as e:
                    logger.error(f'Error parsing cookies: {e}')

            elif event == 'onReceivedHttpError':
                response_data = json.loads(data)
                response = Response(
                    response_data.get('url', ''),
                    response_data.get('statusCode', 0),
                    response_data.get('headers', {}),
                )
                self.pywebview_window.events.response_received.set(response)

        self._cookies = []
        self.webview = WebView(activity)
        webview_settings = self.webview.getSettings()
        webview_settings.setAllowFileAccessFromFileURLs(True)
        webview_settings.setJavaScriptEnabled(True)
        webview_settings.setUseWideViewPort(False)
        webview_settings.setLoadWithOverviewMode(True)
        webview_settings.setSupportZoom(self.pywebview_window.zoomable)
        webview_settings.setBuiltInZoomControls(False)
        webview_settings.setDomStorageEnabled(not _state['private_mode'])

        if _state['user_agent']:
            webview_settings.setUserAgentString(_state['user_agent'])

        self._webview_callback_wrapper = EventCallbackWrapper(webview_callback)
        self._webview_client = webview_client = PyWebViewClient()
        webview_client.setCallback(
            self._webview_callback_wrapper, _state['ssl'] or settings['IGNORE_SSL_ERRORS']
        )
        self._request_interceptor = RequestInterceptor(self._on_request, self._on_response)
        webview_client.setRequestInterceptor(self._request_interceptor)

        self.webview.setWebViewClient(webview_client)

        self._chrome_callback_wrapper = EventCallbackWrapper(chrome_callback)
        self._chrome_client = chrome_client = PyWebChromeClient()
        chrome_client.setCallback(self._chrome_callback_wrapper)
        self.webview.setWebChromeClient(chrome_client)

        self._js_api_callback_wrapper = JsApiCallbackWrapper(js_api_handler)
        self._js_interface = js_interface = PyJavascriptInterface()
        js_interface.setCallback(self._js_api_callback_wrapper)
        self.webview.addJavascriptInterface(js_interface, 'external')

        if settings['ALLOW_DOWNLOADS']:

            def _on_download_start(url, *_):
                context = activity.getApplicationContext()
                visibility = DownloadManagerRequest.VISIBILITY_VISIBLE_NOTIFY_COMPLETED
                dir_type = Environment.DIRECTORY_DOWNLOADS
                uri = Uri.parse(url)
                filepath = uri.getLastPathSegment()
                request = DownloadManagerRequest(uri)
                request.setNotificationVisibility(visibility)
                request.setDestinationInExternalFilesDir(context, dir_type, filepath)
                dm = cast(
                    'android.app.DownloadManager',
                    activity.getSystemService(Context.DOWNLOAD_SERVICE),
                )
                dm.enqueue(request)

            self._download_listener = DownloadListener(_on_download_start)
            self.webview.setDownloadListener(self._download_listener)

        self._key_listener = KeyListener(self._back_pressed)
        self.webview.setOnKeyListener(self._key_listener)
        self.pywebview_window.events.before_show.set()

        if self.pywebview_window.real_url:
            self.webview.loadUrl(self.pywebview_window.real_url)
        elif self.pywebview_window.html:
            self.webview.loadDataWithBaseURL(
                None, self.pywebview_window.html, 'text/html', 'UTF-8', None
            )

        if self.pywebview_window.fullscreen:
            toggle_fullscreen(self.pywebview_window)

        self.pywebview_window.events.shown.set()
        activity.setContentView(self.webview)

    def _on_request(self, url: str, method: str, headers_json: str):
        headers = json.loads(headers_json) if headers_json else {}
        original_headers = headers.copy()

        request = Request(url, method, headers)
        self.pywebview_window.events.request_sent.set(request)

        if request.headers != original_headers:
            logger.debug(
                'Request headers mutated. Original: %s, Mutated: %s',
                original_headers,
                request.headers,
            )

            return json.dumps(request.headers)

        return None

    def _on_response(self, url: str, status_code: int, headers_json: str):
        headers = json.loads(headers_json) if headers_json else {}

        response = Response(url, status_code, headers)
        self.pywebview_window.events.response_received.set(response)

    def dismiss(self):
        @run_on_ui_thread
        def _dismiss():
            try:
                if _state['private_mode']:
                    self.webview.clearHistory()
                    self.webview.clearCache(True)
                    self.webview.clearFormData()

                if hasattr(self, '_js_interface') and self._js_interface:
                    self.webview.removeJavascriptInterface('external')
                    self._js_interface = None
                    logger.error('Removed JavaScript interface')

                if hasattr(self, '_webview_client') and self._webview_client:
                    self._webview_client.destroy()
                    self._webview_client = None

                if hasattr(self, '_webview_callback_wrapper'):
                    self._webview_callback_wrapper = None
                if hasattr(self, '_chrome_callback_wrapper'):
                    self._chrome_callback_wrapper = None

                if hasattr(self, '_js_api_callback_wrapper'):
                    self._js_api_callback_wrapper = None

                if hasattr(self, '_request_interceptor'):
                    self._request_interceptor = None
                    self._download_listener = None

                if hasattr(self, '_key_listener'):
                    self._key_listener = None

                self.webview.destroy()
                self.layout = None
                self.webview = None
            except Exception as e:
                logger.error(f'Error during dismiss: {e}')
            finally:
                app.stop()

        _dismiss()

    def _quit_confirmation(self):
        def cancel(dialog, which):
            self.dialog = None

        def quit(dialog, which):
            self.pywebview_window.closed.set()
            self.dialog = None
            app.stop()

        if self.dialog:
            return

        String = autoclass('java.lang.String')
        message = String(self.pywebview_window.localization['global.quitConfirmation'])
        quit_msg = String(self.pywebview_window.localization['global.quit'])
        cancel_msg = String(self.pywebview_window.localization['global.cancel'])

        self.dialog = (
            AlertDialogBuilder(activity)
            .setMessage(message)
            .setPositiveButton(quit_msg, quit)
            .setNegativeButton(cancel_msg, cancel)
            .setOnCancelListener(cancel)
            .show()
        )

    def _back_pressed(self, v, key_code, event):
        if not (event.getAction() == KeyEvent.ACTION_DOWN and key_code == KeyEvent.KEYCODE_BACK):
            return False
        elif self.webview.canGoBack():
            self.webview.goBack()
        elif self.pywebview_window.events.closing.set():
            pass
        elif self.pywebview_window.confirm_close:
            self._quit_confirmation()
        else:
            app.pause()
            self.pywebview_window.events.closed.set()
        return True

    def get_size(self):
        lock = Semaphore(0)
        size = None, None

        @run_on_ui_thread
        def _get_size():
            nonlocal size
            size = self.webview.getWidth(), self.webview.getHeight()
            lock.release()

        _get_size()
        lock.acquire()

        return size

    def get_url(self):
        lock = Semaphore(0)
        url = None

        @run_on_ui_thread
        def _get_url():
            nonlocal url
            url = self.webview.getUrl()
            lock.release()

        _get_url()
        lock.acquire()

        return url

    @run_on_ui_thread
    def load_url(self, url):
        self.webview.loadUrl(url)

    @run_on_ui_thread
    def load_data_with_base_url(self, base_uri, html_content):
        self.webview.loadDataWithBaseURL(base_uri, html_content, 'text/html', 'UTF-8', None)


class AndroidApp(App):
    def __init__(self, window):
        self.window = window
        self.first_show = True
        super().__init__()

        self.register_event_type('on_create')
        self.register_event_type('on_destroy')
        self.register_event_type('on_pause')
        self.register_event_type('on_resume')
        self.register_event_type('on_start')
        self.register_event_type('on_stop')

    def build_view(self):
        self.view = BrowserView(self.window)
        return self.view

    def on_pause(self, _):
        self.view.webview.pauseTimers()
        self.view.webview.onPause()

        logger.debug('pausing initiated')

    def on_resume(self, _):
        self.view.webview.onResume()
        self.view.webview.resumeTimers()

    def on_create(self, activity, saved_instance_state):
        """Event handler for the `on_create` event which is fired after
        initialization (after build() has been called) but before the
        application has started running.
        """

    def on_destroy(self, _):
        self.view.dismiss()

    def on_start(self, activity):
        """Event handler for the `on_start` event which is fired when the
        application is becoming visible to the user.
        """

    def on_stop(self, activity):
        """Event handler for the `on_stop` event which is fired when the
        application is no longer visible to the user.
        """


def create_file_dialog(*_):
    logger.warning('Creating file dialogs is not supported on Android')


def create_window(window):
    global app

    if app:
        logger.error('Multiple windows are not supported on Android')
        return

    app = AndroidApp(window)
    app.run()


def setup_app():
    pass


def load_url(url, _):
    app.view._cookies = []
    app.view.load_url(url)


def load_html(html_content, base_uri, _):
    app.view._cookies = []
    app.view.load_data_with_base_url(base_uri, html_content)


def evaluate_js(js_code, _, parse_json=True):
    def callback(result):
        nonlocal js_result, value_callback
        try:
            # The result is double-encoded in Android, once by the WebView and once pywebview's stringify
            js_result = json.loads(result) if result else result
            js_result = json.loads(js_result) if parse_json and js_result else js_result
        except Exception as e:
            logger.exception(f'Error parsing result: {js_result}. Type: {type(js_result)}\n{e}')
        finally:
            value_callback = None
            lock.release()

    @run_on_ui_thread
    def _evaluate_js():
        nonlocal value_callback
        try:
            value_callback = ValueCallback(callback)
            app.view.webview.evaluateJavascript(js_code, value_callback)
        except Exception as e:
            logger.error(f'Error evaluating JavaScript: {e}')
            # Ensure callback is cleared on error
            value_callback = None
            lock.release()

    lock = Semaphore(0)
    js_result = None
    value_callback = None

    _evaluate_js()
    lock.acquire()
    return js_result


def clear_cookies(_):
    lock = Semaphore(0)

    @run_on_ui_thread
    def _clear_cookies():
        cookie_manager = None
        try:
            # Get fresh CookieManager reference to avoid stale Java object references
            cookie_manager = CookieManager.getInstance()
            cookie_manager.removeAllCookies(None)
        except Exception as e:
            logger.error(f'Error clearing cookies: {e}')
        finally:
            # Clear the reference to help GC
            cookie_manager = None
            lock.release()

    app.view._cookies = []
    _clear_cookies()
    lock.acquire()


def get_cookies(_):
    cookies = app.view._cookies
    cookie_string = evaluate_js('document.cookie', None, False)

    try:
        current_url = app.view.get_url()

        # Parse URL once outside the loop to avoid repeated parsing
        parsed_url = urlparse(current_url)
        domain = parsed_url.netloc
        is_secure = current_url.startswith('https')

        # Parse the cookie string into individual cookies
        for cookie_pair in cookie_string.split(';'):
            cookie_pair = cookie_pair.strip()
            if '=' in cookie_pair:
                name, value = cookie_pair.split('=', 1)
                name = name.strip()

                existing_cookie = next((c for c in cookies if name in c), None)

                if existing_cookie and existing_cookie[name].value == value.strip():
                    # If the cookie already exists, skip it
                    continue

                cookie_dict = {
                    'name': name,
                    'expires': None,  # Android does not provide expiration in getCookie
                    'value': value.strip(),
                    'domain': domain,
                    'path': '/',
                    'secure': is_secure,
                    'httponly': False,
                }

                cookies.append(create_cookie(cookie_dict))
    except Exception as e:
        logger.exception(f'Error getting cookies: {e}')

    return cookies


def get_current_url(_):
    lock = Semaphore(0)
    url = None

    @run_on_ui_thread
    def _get_current_url():
        nonlocal url
        try:
            url = app.view.get_url()
        except Exception as e:
            logger.error(f'Error getting current URL: {e}')
        finally:
            lock.release()

    _get_current_url()
    lock.acquire()
    return url


def get_screens():
    logger.warning('Screen information is not supported on Android')
    return []


def get_size(_):
    return app.view.get_size()


def get_position(_):
    return (0, 0)


def show(_):
    logger.warning('Showing window is not supported on Android')


def hide(_):
    app.pause()


def minimize(_):
    app.pause()


def move(_):
    logger.warning('Moving window is not supported on Android')


def restore(_):
    logger.warning('Restoring window is not supported on Android')


def resize(width, height, _, fix_point):
    logger.warning('Resizing window is not supported on Android')


def destroy_window():
    app.stop()


def set_title(title, _):
    logger.warning('Changing app title is not supported on Android')


def set_on_top(_, on_top):
    logger.warning('Always on top mode is not supported on Android')


@run_on_ui_thread
def toggle_fullscreen(_):
    is_fullscreen = app.view.is_fullscreen

    try:
        if not is_fullscreen:
            option = (
                View.SYSTEM_UI_FLAG_FULLSCREEN
                | View.SYSTEM_UI_FLAG_HIDE_NAVIGATION
                | View.SYSTEM_UI_FLAG_IMMERSIVE_STICKY
                | View.SYSTEM_UI_FLAG_LAYOUT_STABLE
                | View.SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION
                | View.SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN
            )
            app.view.webview.setSystemUiVisibility(option)
            app.view.is_fullscreen = True
        else:
            option = View.SYSTEM_UI_FLAG_VISIBLE
            app.view.webview.setSystemUiVisibility(option)
            app.view.is_fullscreen = False
    except Exception as e:
        logger.error(f'Error toggling fullscreen: {e}')


def add_tls_cert(certfile):
    pass



"""
All Java interfaces should go to this package.
"""

from webview.platforms.android.jinterface.pywebview import (
    EventCallbackWrapper,
    JsApiCallbackWrapper,
    RequestInterceptor,
)
from webview.platforms.android.jinterface.view import KeyListener
from webview.platforms.android.jinterface.webkit import DownloadListener, ValueCallback

__all__ = (
    'EventCallbackWrapper',
    'JsApiCallbackWrapper',
    'RequestInterceptor',
    'KeyListener',
    'ValueCallback',
    'DownloadListener',
)



from PyInstaller.compat import is_win
from PyInstaller.utils.hooks import collect_data_files, collect_dynamic_libs

datas = []
if is_win:
    datas = collect_data_files('webview', subdir='lib')
    binaries = collect_dynamic_libs('webview')


datas += collect_data_files('webview', subdir='js')



import ctypes
import logging as _logging
from ctypes import wintypes

_log = _logging.getLogger('pywebview')


_user32 = ctypes.windll.user32


# DEVMODE structure for EnumDisplaySettings
class DEVMODE(ctypes.Structure):
    _fields_ = [
        ('dmDeviceName', wintypes.WCHAR * 32),
        ('dmSpecVersion', wintypes.WORD),
        ('dmDriverVersion', wintypes.WORD),
        ('dmSize', wintypes.WORD),
        ('dmDriverExtra', wintypes.WORD),
        ('dmFields', wintypes.DWORD),
        ('dmPositionX', wintypes.LONG),
        ('dmPositionY', wintypes.LONG),
        ('dmDisplayOrientation', wintypes.DWORD),
        ('dmDisplayFixedOutput', wintypes.DWORD),
        ('dmColor', wintypes.SHORT),
        ('dmDuplex', wintypes.SHORT),
        ('dmYResolution', wintypes.SHORT),
        ('dmTTOption', wintypes.SHORT),
        ('dmCollate', wintypes.SHORT),
        ('dmFormName', wintypes.WCHAR * 32),
        ('dmLogPixels', wintypes.WORD),
        ('dmBitsPerPel', wintypes.DWORD),
        ('dmPelsWidth', wintypes.DWORD),
        ('dmPelsHeight', wintypes.DWORD),
        ('dmDisplayFlags', wintypes.DWORD),
        ('dmDisplayFrequency', wintypes.DWORD),
        ('dmICMMethod', wintypes.DWORD),
        ('dmICMIntent', wintypes.DWORD),
        ('dmMediaType', wintypes.DWORD),
        ('dmDitherType', wintypes.DWORD),
        ('dmReserved1', wintypes.DWORD),
        ('dmReserved2', wintypes.DWORD),
        ('dmPanningWidth', wintypes.DWORD),
        ('dmPanningHeight', wintypes.DWORD),
    ]


# DISPLAY_DEVICEW structure for EnumDisplayDevices
class DISPLAY_DEVICEW(ctypes.Structure):
    _fields_ = [
        ('cb', wintypes.DWORD),
        ('DeviceName', wintypes.WCHAR * 32),
        ('DeviceString', wintypes.WCHAR * 128),
        ('StateFlags', wintypes.DWORD),
        ('DeviceID', wintypes.WCHAR * 128),
        ('DeviceKey', wintypes.WCHAR * 128),
    ]


def get_screen_scale(device_name: str, logical_width: int, logical_height: int) -> float:
    """
    Calculate the DPI scale factor for a screen.

    Args:
        device_name: The device name (e.g., "\\\\.\\DISPLAY1")
        logical_width: The logical width in pixels (DPI-scaled)
        logical_height: The logical height in pixels (DPI-scaled)

    Returns:
        The scale factor (e.g., 1.0, 1.5, 2.0, etc.)
    """
    try:
        dm = DEVMODE()
        dm.dmSize = ctypes.sizeof(DEVMODE)

        # ENUM_CURRENT_SETTINGS = -1
        if _user32.EnumDisplaySettingsW(device_name, -1, ctypes.byref(dm)):
            physical_width = dm.dmPelsWidth
        else:
            # Fallback to logical size if EnumDisplaySettings fails
            return 1.0

        # Calculate scale from the ratio
        if logical_width > 0 and logical_height > 0:
            return physical_width / logical_width
        else:
            return 1.0

    except Exception as e:
        _log.debug(f'Failed to get display settings: {e}')
        return 1.0



from typing import Dict

original_localization: Dict[str, str] = {
    'global.quitConfirmation': 'Do you really want to quit?',
    'global.ok': 'OK',
    'global.quit': 'Quit',
    'global.cancel': 'Cancel',
    'global.saveFile': 'Save file',
    'cocoa.menu.about': 'About',
    'cocoa.menu.services': 'Services',
    'cocoa.menu.view': 'View',
    'cocoa.menu.edit': 'Edit',
    'cocoa.menu.hide': 'Hide',
    'cocoa.menu.hideOthers': 'Hide Others',
    'cocoa.menu.showAll': 'Show All',
    'cocoa.menu.quit': 'Quit',
    'cocoa.menu.fullscreen': 'Enter Fullscreen',
    'cocoa.menu.cut': 'Cut',
    'cocoa.menu.copy': 'Copy',
    'cocoa.menu.paste': 'Paste',
    'cocoa.menu.selectAll': 'Select All',
    'windows.fileFilter.allFiles': 'All files',
    'windows.fileFilter.otherFiles': 'Other file types',
    'linux.openFile': 'Open file',
    'linux.openFiles': 'Open files',
    'linux.openFolder': 'Open folder',
}



__all__ = ('AlertDialogBuilder', 'DownloadManagerRequest')

from jnius import JavaClass, JavaMethod, JavaMultipleMethod, JavaStaticField, MetaJavaClass


class AlertDialogBuilder(JavaClass, metaclass=MetaJavaClass):
    """
    Represents a builder for creating AlertDialog instances in the Android platform.

    This class provides a fluent API for constructing and customizing AlertDialog instances
    with various settings such as title, message, icon, buttons, custom views, and other
    behavioral options. It serves as a wrapper around the native Java AlertDialog.Builder
    implementation, enabling its usage within Python applications interfacing with Android.

    Attributes:
        __javaclass__ (str): Specifies the Java class represented by this Python class.
        __javaconstructor__ (list): Defines available constructors in the underlying Java class.

    Methods:
        setTitle: Sets the title of the AlertDialog.
        setCustomTitle: Sets a custom view to be used as the title of the AlertDialog.
        setMessage: Specifies the message text for the AlertDialog.
        setIcon: Sets the icon for the AlertDialog.
        setIconAttribute: Sets an attribute as the icon for the AlertDialog.
        setPositiveButton: Sets positive button text and its click listener.
        setNegativeButton: Sets negative button text and its click listener.
        setNeutralButton: Sets neutral button text and its click listener.
        setCancelable: Configures whether the AlertDialog is cancelable or not.
        setOnCancelListener: Sets the cancel listener to invoke when the AlertDialog is canceled.
        setOnDismissListener: Sets a listener for when the AlertDialog is dismissed.
        setOnKeyListener: Sets a key listener for key events within the AlertDialog.
        setItems: Populates the AlertDialog with a list of selectable items.
        setAdapter: Populates the AlertDialog with a list using a custom adapter.
        setCursor: Sets a cursor to define the data to be displayed in the AlertDialog.
        setMultiChoiceItems: Configures multiple choice items and their states for the AlertDialog.
        setSingleChoiceItems: Configures single-choice items and their states for the AlertDialog.
        setView: Sets a custom view to be displayed in the AlertDialog.
        setInverseBackgroundForced: Specifies whether to force the AlertDialog to use an inverse background.
        create: Constructs an AlertDialog instance without displaying it.
        show: Shows the AlertDialog after construction.
    """

    __javaclass__ = 'android/app/AlertDialog$Builder'
    __javaconstructor__ = [
        ('(Landroid/content/Context;)V', False),
        ('(Landroid/content/Context;I)V', False),
    ]
    setTitle = JavaMultipleMethod(
        [
            ('(I)Landroid/app/AlertDialog$Builder;', False, False),
            ('(Ljava/lang/CharSequence;)Landroid/app/AlertDialog$Builder;', False, False),
        ]
    )
    setCustomTitle = JavaMethod('(Landroid/view/View;)Landroid/app/AlertDialog$Builder;')
    setMessage = JavaMultipleMethod(
        [
            ('(I)Landroid/app/AlertDialog$Builder;', False, False),
            ('(Ljava/lang/CharSequence;)Landroid/app/AlertDialog$Builder;', False, False),
        ]
    )
    setIcon = JavaMultipleMethod(
        [
            ('(I)Landroid/app/AlertDialog$Builder;', False, False),
            (
                '(Landroid/graphics/drawable/Drawable;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setIconAttribute = JavaMethod('(I)Landroid/app/AlertDialog$Builder;')
    setPositiveButton = JavaMultipleMethod(
        [
            (
                '(ILandroid/content/DialogInterface$OnClickListener;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Ljava/lang/CharSequence;Landroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setNegativeButton = JavaMultipleMethod(
        [
            (
                '(ILandroid/content/DialogInterface$OnClickListener;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Ljava/lang/CharSequence;Landroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setNeutralButton = JavaMultipleMethod(
        [
            (
                '(ILandroid/content/DialogInterface$OnClickListener;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Ljava/lang/CharSequence;Landroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setCancelable = JavaMethod('(Z)Landroid/app/AlertDialog$Builder;')
    setOnCancelListener = JavaMethod(
        '(Landroid/content/DialogInterface$OnCancelListener;)Landroid/app/AlertDialog$Builder;'
    )
    setOnDismissListener = JavaMethod(
        '(Landroid/content/DialogInterface$OnDismissListener;)Landroid/app/AlertDialog$Builder;'
    )
    setOnKeyListener = JavaMethod(
        '(Landroid/content/DialogInterface$OnKeyListener;)Landroid/app/AlertDialog$Builder;'
    )
    setItems = JavaMultipleMethod(
        [
            (
                '(ILandroid/content/DialogInterface$OnClickListener;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '([Ljava/lang/CharSequence;Landroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setAdapter = JavaMethod(
        '(Landroid/widget/ListAdapter;Landroid/content/DialogInterface$OnClickListener;)'
        'Landroid/app/AlertDialog$Builder;'
    )
    setCursor = JavaMethod(
        '(Landroid/database/Cursor;Landroid/content/DialogInterface$OnClickListener;'
        'Ljava/lang/String;)Landroid/app/AlertDialog$Builder;'
    )
    setMultiChoiceItems = JavaMultipleMethod(
        [
            (
                '(I[ZLandroid/content/DialogInterface$OnMultiChoiceClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '([Ljava/lang/CharSequence;[ZLandroid/content/DialogInterface$OnMultiChoiceClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Landroid/database/Cursor;Ljava/lang/String;'
                'Ljava/lang/String;Landroid/content/DialogInterface$OnMultiChoiceClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
        ]
    )
    setSingleChoiceItems = JavaMultipleMethod(
        [
            (
                '(IILandroid/content/DialogInterface$OnClickListener;)Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Landroid/database/Cursor;ILjava/lang/String;Landroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '([Ljava/lang/CharSequence;ILandroid/content/DialogInterface$OnClickListener;)'
                'Landroid/app/AlertDialog$Builder;',
                False,
                False,
            ),
            (
                '(Landroid/widget/ListAdapter;ILandroid/content/DialogInterface$OnClickListener;)',
                False,
                False,
            ),
        ]
    )
    setView = JavaMultipleMethod(
        [
            ('(I)Landroid/app/AlertDialog$Builder;', False, False),
            ('(Landroid/view/View;)Landroid/app/AlertDialog$Builder;', False, False),
        ]
    )
    setInverseBackgroundForced = JavaMethod('(Z)Landroid/app/AlertDialog$Builder;')
    create = JavaMethod('()Landroid/app/AlertDialog;')
    show = JavaMethod('()Landroid/app/AlertDialog;')


class DownloadManagerRequest(JavaClass, metaclass=MetaJavaClass):
    """
    Represents a request for downloading a file, providing methods to configure
    download behavior and properties. This class is a part of the Android Download
    Manager system, enabling developers to easily handle file downloads in
    background.

    The DownloadManagerRequest class allows setting various options for a download
    request, including destination, network restrictions, visibility, and more. It is
    designed to be used as a configuration object passed to the DownloadManager
    to initiate a download task.

    Attributes:
        NETWORK_MOBILE: int
            Specifies that downloads are allowed over a mobile network connection.
        NETWORK_WIFI: int
            Specifies that downloads are allowed over a Wi-Fi network connection.
        VISIBILITY_HIDDEN: int
            Visibility state where the download is hidden from the user interface.
        VISIBILITY_VISIBLE: int
            Visibility state where the download is visible in the user interface.
        VISIBILITY_VISIBLE_NOTIFY_COMPLETED: int
            Visibility state where the download's completion will notify the user.
        VISIBILITY_VISIBLE_NOTIFY_ONLY_COMPLETION: int
            Visibility state where only completion of the download notifies the user.
    """

    __javaclass__ = 'android/app/DownloadManager$Request'
    __javaconstructor__ = [('(Landroid/net/Uri;)V', False)]
    NETWORK_MOBILE = JavaStaticField('I')
    NETWORK_WIFI = JavaStaticField('I')
    VISIBILITY_HIDDEN = JavaStaticField('I')
    VISIBILITY_VISIBLE = JavaStaticField('I')
    VISIBILITY_VISIBLE_NOTIFY_COMPLETED = JavaStaticField('I')
    VISIBILITY_VISIBLE_NOTIFY_ONLY_COMPLETION = JavaStaticField('I')
    setDestinationUri = JavaMethod('(Landroid/net/Uri;)Landroid/app/DownloadManager$Request;')
    setDestinationInExternalFilesDir = JavaMethod(
        '(Landroid/content/Context;Ljava/lang/String;Ljava/lang/String;)'
        'Landroid/app/DownloadManager$Request;'
    )
    setDestinationInExternalPublicDir = JavaMethod(
        '(Ljava/lang/String;Ljava/lang/String;)Landroid/app/DownloadManager$Request;'
    )
    allowScanningByMediaScanner = JavaMethod('()V')
    addRequestHeader = JavaMethod(
        '(Ljava/lang/String;Ljava/lang/String;)Landroid/app/DownloadManager$Request;'
    )
    setTitle = JavaMethod('(Ljava/lang/CharSequence;)Landroid/app/DownloadManager$Request;')
    setDescription = JavaMethod('(Ljava/lang/CharSequence;)Landroid/app/DownloadManager$Request;')
    setMimeType = JavaMethod('(Ljava/lang/String;)Landroid/app/DownloadManager$Request;')
    setShowRunningNotification = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
    setNotificationVisibility = JavaMethod('(I)Landroid/app/DownloadManager$Request;')
    setAllowedNetworkTypes = JavaMethod('(I)Landroid/app/DownloadManager$Request;')
    setAllowedOverRoaming = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
    setAllowedOverMetered = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
    setRequiresDeviceIdle = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
    setRequiresCharging = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
    setVisibleInDownloadsUi = JavaMethod('(Z)Landroid/app/DownloadManager$Request;')
